import sys
from snakemake.shell import shell

class DependencyChecker:

    def __init__(self, tool, version=None):
        self.tool = tool
        self.version = version

    def check(self, cmd='{tool} --version 2>&1', extract_version= lambda x: x):
        self.check_exists()
        output = shell(cmd.format(tool=self.tool), read=True).decode('utf-8').strip()
        version = extract_version(output)
        print("Current version of {tool}: {version}".format(tool=self.tool, version=version))
        print("Expected version of {tool}: {version}".format(tool=self.tool, version=self.version))
        if version != self.version:
            print("WARNING: Current version of {tool} does not match expectaction. Proceed with caution".format(tool=self.tool))


    def check_exists(self):
        try:
            path = shell('which {tool}'.format(tool=self.tool), read=True)
        except:
            print("MISSING DEPENDENCY: {tool} is not installed. Please install before continuing.".format(tool=self.tool))
            sys.exit()
            


def check_dependencies():

    print("#### CHECKING DEPENDENCIES ####")
    snakemake_checker = DependencyChecker('snakemake', '3.13.3')
    snakemake_checker.check()

    emboss_checker = DependencyChecker('einverted', 'EMBOSS:6.6.0.0')
    emboss_checker.check()

    bowtie2_checker = DependencyChecker('bowtie2', '2.3.5')
    bowtie2_checker.check(extract_version=lambda x: x.split()[2])

    samtools_checker = DependencyChecker('samtools', '1.9')
    samtools_checker.check(extract_version=lambda x: x.split()[1])

    cdhit_checker = DependencyChecker('cd-hit', '4.8.1')
    cdhit_checker.check(cmd='{tool} || exit 0', extract_version=lambda x: x.split()[3])
    print("###############################")



import warnings
warnings.filterwarnings("ignore")
from snakemake import shell
from os.path import isfile

def remove_secondary_alignments(in_bam, out_bam, delete_in_bam=False):
    shell('samtools view -b -h -F 0x900 {in_bam} > {out_bam}'.format(in_bam=in_bam, out_bam=out_bam))

    if delete_in_bam:
        shell('rm {in_bam}'.format(in_bam=in_bam))

    if isfile(out_bam):
        return True
    else:
        return False


def sort_coordinate(in_bam, out_bam, delete_in_bam=False):
    shell('samtools sort {in_bam} > {out_bam}'.format(in_bam=in_bam, out_bam=out_bam))

    if delete_in_bam:
        shell('rm {in_bam}'.format(in_bam=in_bam))

    if isfile(out_bam):
        return True
    else:
        return False

def index(in_bam):
    shell('samtools index {in_bam}'.format(in_bam=in_bam))

    if isfile(in_bam+'.bai'):
        return True
    else:
        return False


from glob import glob
from snakemake import shell
from os.path import isfile

def index_genome(genome_path, silence=True):

    if silence:
        shell('makeblastdb -dbtype nucl -in {genome_path} &> /dev/null;'.format(genome_path=genome_path))
    else:
        shell('makeblastdb -dbtype nucl -in {genome_path}'.format(genome_path=genome_path))

    return genome_is_indexed(genome_path)


def genome_is_indexed(genome_path):

    nhr = genome_path + '.nhr'
    nin = genome_path + '.nin'
    nsq = genome_path + '.nsq'

    index_files = [nhr, nin, nsq]

    indexed = True
    files = glob(genome_path+'*')
    for f in index_files:
        if f not in files:
            indexed = False

    return indexed


def align_fasta_to_genome(fasta, genome_path, outfile, threads=1, silence=True, additional_flags=''):

    shell('blastn -query {fasta} -db {genome_path} -outfmt 5 -max_target_seqs 100000 -out {outfile} -parse_deflines '
          '{additional_flags}'.format(
        fasta=fasta, genome_path=genome_path, outfile=outfile, additional_flags=additional_flags
    ))

    if isfile(outfile):
        return True
    else:
        return False


import sys
from snakemake import shell
try:
    from snakemake import shell
except ImportError:
    print("MISSING DEPENDENCY: Snakemake is not installed. You can install with the command \"conda install -c bioconda -c conda-forge snakemake\" or \"pip3 install snakemake\"")
    sys.exit()



def _workflow(workdir, snakefile, configfile, cores, memory, unlock, rerun_incomplete, keep_going):

    if unlock == True:
        cmd = 'snakemake -s {snakefile} --configfile {configfile} --config wd={workdir} --unlock'.format(
              snakefile=snakefile, configfile=configfile, workdir=workdir
        )
    cmd = 'snakemake -s {snakefile} --config wd={workdir} memory={memory} ' \
          '--cores {cores} --configfile {configfile} '
    if rerun_incomplete:
        cmd += '--rerun-incomplete '
    if keep_going:
        cmd += '--keep-going '
    
    cmd = cmd.format(snakefile=snakefile, configfile=configfile, workdir=workdir, memory=memory, cores=cores)
    print('COMMAND:', cmd)
    shell(cmd)



import warnings
warnings.filterwarnings("ignore")
import click
import pandas as pd
import numpy as np
from snakemake import shell
from random import randint
from mgefinder import fastatools, embosstools, pysamtools, sctools
from os.path import basename, join, dirname
from Bio import SeqIO
import pysam
from collections import defaultdict


def _pair(findfile, bamfile, genome, max_direct_repeat_length, min_alignment_quality,
                min_alignment_inner_length, max_junction_spanning_prop=0.15, large_insertion_cutoff=30,
                output_file=None):
    tmp_output_prefix = '.'.join(basename(output_file).split('.')[:-1])

    flanks = pd.read_csv(findfile, sep='\t')


    bam = pysam.AlignmentFile(bamfile, 'rb')

    flank_pairer = FlankPairer(flanks, bam, genome, max_direct_repeat_length, min_alignment_quality,
                               min_alignment_inner_length, max_junction_spanning_prop, large_insertion_cutoff,
                               tmp_dir=dirname(output_file), tmp_output_prefix=tmp_output_prefix)

    if flanks.shape[0] == 0:
        click.echo("No flanks found in the input file...")

        flank_pairs = flank_pairer.get_header_dataframe()
        flank_pairs.insert(0, 'sample', None)

        if output_file:
            flank_pairs.to_csv(output_file, sep='\t', index=False)
        return flank_pairs

    else:
        flank_pairs = flank_pairer.run_pair_flanks()
        sample_id = list(flanks['sample'])[0]
        flank_pairs.insert(0, 'sample', sample_id)

        if output_file:
            click.echo("Saving results to file %s" % output_file)
            flank_pairs.to_csv(output_file, sep='\t', index=False)

        return flank_pairs


class FlankPairer:

    flanks = None
    bam = None
    genome = None
    max_direct_repeat_length = None
    min_alignment_quality = None
    min_alignment_inner_length = None
    max_junction_spanning_prop = None
    truncated_flank_length = None
    ir_distance_from_end = None

    insertion_spanning_length = None
    large_insertion_cutoff = None

    tmp_dir = None
    tmp_output_prefix = None

    flank_pairs = None


    def __init__(self, flanks, bam, genome, max_direct_repeat_length,
                 min_alignment_quality, min_alignment_inner_length, max_junction_spanning_prop,
                 large_insertion_cutoff,
                 truncated_flank_length=40, ir_distance_from_end=15, insertion_spanning_length=10,
                 tmp_dir='/tmp', tmp_output_prefix='mgefinder'):
        self.flanks = flanks
        self.bam = bam
        self.genome = genome
        self.max_direct_repeat_length = max_direct_repeat_length
        self.min_alignment_quality = min_alignment_quality
        self.min_alignment_inner_length = min_alignment_inner_length
        self.max_junction_spanning_prop = max_junction_spanning_prop
        self.large_insertion_cutoff = large_insertion_cutoff
        self.truncated_flank_length = truncated_flank_length
        self.ir_distance_from_end = ir_distance_from_end
        self.insertion_spanning_length = insertion_spanning_length

        self.tmp_dir = tmp_dir
        self.tmp_output_prefix = tmp_output_prefix


    def run_pair_flanks(self):
        click.echo("Finding all flank pairs within %d bases of each other ..." % self.max_direct_repeat_length)
        pairs = self.pair_all_nearby_flanks(self.flanks)
        click.echo("Finding all inverted repeats at termini in %d candidate pairs..." % pairs.shape[0])
        pairs = self.check_pairs_for_ir(pairs)
        click.echo("Assigning pairs according to existence of inverted repeats, read count difference, and flank length difference...")
        assigned_pairs = self.assign_pairs(pairs)
        click.echo("Filtering out pairs with evidence of reads spanning both clipped junctions...")
        analyzed_pairs = self.count_insertion_spanning_reads(assigned_pairs)

        click.echo("Identified %d flank pairs in total..." % analyzed_pairs.shape[0])
        click.echo("Identified %d flank pairs with inverted repeats..." % analyzed_pairs.query('has_IR==True').shape[0])
        click.echo("Identified %d flank pairs with reads that span the insertion..." % analyzed_pairs.query('spanning_count > 0').shape[0])

        click.echo("Filtering sites with junction-spanning reads...")
        filtered_pairs = self.filter_junction_spanning(analyzed_pairs)
        click.echo("%d flank pairs remain after filtering..." % filtered_pairs.shape[0])


        click.echo("Getting direct repeats and surrounding genomic region...")
        final_pairs = self.get_direct_repeats(filtered_pairs)



        final_pairs['pair_id'] = list(range(1, final_pairs.shape[0]+1))

        return final_pairs


    def get_header_list(self):
        header = ['pair_id', 'contig', 'pos_5p', 'pos_3p', 'softclip_count_5p', 'softclip_count_3p',
                  'total_count_5p', 'total_count_3p', 'spanning_count', 'has_IR', 'IR_length', 'IR_5p', 'IR_3p',
                  'seq_5p', 'seq_3p',  'direct_repeat_reference']

        return header

    def get_header_dataframe(self):
        flank_pairs = pd.DataFrame(columns=self.get_header_list())

        return flank_pairs


    def pair_all_nearby_flanks(self, flanks):

        column_names = ['contig', 'index_5p', 'index_3p', 'pos_5p', 'pos_3p', 'softclip_count_5p', 'softclip_count_3p',
                        'total_count_5p', 'total_count_3p']

        column_names += ['seq_5p', 'seq_3p']

        outpairs = dict()

        for index_5p, row in flanks.iterrows():

            if row.orient != '5p':
                continue

            contig, pos_5p = str(row['contig']), row['pos']
            softclip_count_5p, total_count_5p, seq_5p = row['consensus_softclip_count'], row['total_count'], row['consensus_seq']

            min_pos = pos_5p - self.max_direct_repeat_length - 1

            candidate_pairs = flanks.query('contig == @contig & pos >= @min_pos & pos < @pos_5p & orient == "3p"')

            for index_3p, row2 in candidate_pairs.iterrows():
                pos_3p, softclip_count_3p, total_count_3p, seq_3p = row2['pos'], row2['consensus_softclip_count'], \
                                                                         row2['total_count'], row2['consensus_seq']

                out = [contig, index_5p, index_3p, pos_5p, pos_3p, softclip_count_5p, softclip_count_3p, total_count_5p, total_count_3p]

                out += [seq_5p, seq_3p]
                outpairs[len(outpairs)] =  out

        outpairs = pd.DataFrame.from_dict(outpairs, orient='index', columns=column_names)
        return outpairs


    def check_pairs_for_ir(self, pairs):

        has_ir_all = []
        ir_5p_all = []
        ir_3p_all = []

        for index, row in pairs.iterrows():

            contig, index_5p, index_3p, pos_5p, pos_3p, softclip_count_5p, softclip_count_3p, \
            runthrough_count_5p, runthrough_count_3p, seq_5p, seq_3p = row

            trunc_seq_5p = self.truncate_sequence(seq_5p, self.truncated_flank_length, orient='5p')
            trunc_seq_3p = self.truncate_sequence(seq_3p, self.truncated_flank_length, orient='3p')

            combined_seq = str('N'*20).join([trunc_seq_5p, trunc_seq_3p])


            tmp_fasta_path = join(self.tmp_dir, self.tmp_output_prefix + '.' + str(randint(0, 1e20)) + '.fasta')
            fastatools.write_sequences_to_fasta([combined_seq], tmp_fasta_path)

            tmp_einverted_outfile = join(self.tmp_dir, self.tmp_output_prefix + '.' + str(randint(0, 1e20)) + '.out')
            tmp_einverted_outseq = join(self.tmp_dir, self.tmp_output_prefix + '.' + str(randint(0, 1e20)) + '.fa')
            embosstools.run_einverted(tmp_fasta_path, outfile=tmp_einverted_outfile, outseq=tmp_einverted_outseq)

            has_ir = False
            ir_length = 0
            keep_ir1 = None
            keep_ir2 = None
            for ir1, ir2 in embosstools.read_emboss_seq_results(tmp_einverted_outseq):
                if self.pair_has_ir(ir1, ir2, self.ir_distance_from_end, len(combined_seq)):
                    has_ir = True
                    if len(ir1.seq) > ir_length:
                        keep_ir1 = ir1.seq
                        keep_ir2 = ir2.seq

            shell('rm -f %s' % tmp_fasta_path)
            shell('rm -f %s' % tmp_einverted_outfile)
            shell('rm -f %s' % tmp_einverted_outseq)

            has_ir_all.append(has_ir)
            ir_3p_all.append(keep_ir1)
            ir_5p_all.append(keep_ir2)

        pairs['has_IR'] = has_ir_all
        pairs['IR_5p'] = ir_5p_all
        pairs['IR_3p'] = ir_3p_all

        return pairs


    def assign_pairs(self, pairs):
        pairs.loc[:, 'direct_repeat_length'] = pairs.loc[:, 'pos_5p'] - pairs.loc[:, 'pos_3p'] -1

        pairs['IR_length'] = np.array([len(seq) if seq is not None else 0 for seq in list(pairs.loc[:,'IR_5p'])])
        pairs['diffcount'] = abs(pairs['softclip_count_5p'] - pairs['softclip_count_3p'])
        pairs['difflength'] = abs(np.array(list(map(len, pairs['seq_5p']))) - np.array(list(map(len, pairs['seq_3p']))))
        pairs['ignore_pair'] = False
        sorted_pairs = pairs.sort_values(['IR_length', 'diffcount', 'difflength'], ascending=[False, True, True])

        keep_row = []

        for index in sorted_pairs.index:
            row = sorted_pairs.loc[index, :]

            if not row['ignore_pair']:
                keep_row.append(True)
                index_5p, index_3p = row['index_5p'], row['index_3p']
                has_pair_member = sorted_pairs.query("index_5p == @index_5p | index_3p == @index_3p")
                has_pair_member_indices = has_pair_member.query("index_5p != @index_5p | index_3p != @index_3p").index.values
                sorted_pairs.loc[has_pair_member_indices, 'ignore_pair'] = True
            else:
                keep_row.append(False)

        sorted_pairs.loc[:, 'keep_pair'] = np.array(keep_row)

        assigned_pairs = sorted_pairs.query("keep_pair == True").loc[:,
                         self.get_header_list()].sort_values(['contig', 'pos_5p', 'pos_3p'])

        return assigned_pairs


    def pair_has_ir(self, ir1, ir2, ir_distance_from_end, seqlen):
        if self.ir_near_5prime_end(ir1, ir_distance_from_end) and self.ir_near_3prime_end(ir2, ir_distance_from_end, seqlen):
            return True
        return False

    def ir_near_5prime_end(self, ir1, ir_distance_from_end):
        if ir1.ir_pos_5p <= ir_distance_from_end:
            return True
        return False

    def ir_near_3prime_end(self, ir2, ir_distance_from_end, seqlen):
        if ir2.ir_pos_3p >= (seqlen - ir_distance_from_end):
            return True
        return False

    def truncate_sequence(self, seq, truncated_seq_length, orient='5p'):
        truncated_seq = seq
        if len(truncated_seq) > truncated_seq_length:
            if orient == '5p':
                truncated_seq = truncated_seq[:truncated_seq_length]
            elif orient == '3p':
                truncated_seq = truncated_seq[-truncated_seq_length:]
        return truncated_seq

    def get_direct_repeats(self, flank_pairs):

        genome_dict = {rec.id: rec.seq for rec in SeqIO.parse(self.genome, 'fasta')}
        positions = self.get_reference_direct_repeats(flank_pairs, genome_dict)

        flank_pairs = flank_pairs.drop(['direct_repeat_reference'], axis=1).merge(positions, how='left')
 
        return flank_pairs


    def get_reference_direct_repeats(self, flank_pairs, genome_dict, target_region_size=50):
        positions = flank_pairs.loc[:, ['contig', 'pos_5p', 'pos_3p']].drop_duplicates().reset_index(drop=True)

        direct_repeats = []
        for index, row in positions.iterrows():
            contig, start, end = str(row['contig']), row['pos_3p'], row['pos_5p']

            direct_repeat = genome_dict[contig][(start+1):end]
            direct_repeats.append(''.join(direct_repeat))

        positions['direct_repeat_reference'] = direct_repeats
        return positions


    def count_insertion_spanning_reads(self, assigned_pairs):

        spanning_reads = []
        contig_lengths = pysamtools.get_bam_contig_dict(self.bam)
        for index, row in assigned_pairs.iterrows():
            spanning_read_names = set()
            contig, pos_3p, pos_5p = str(row['contig']), row['pos_3p'], row['pos_5p']
            reads = self.get_reads_at_site(contig, pos_3p, pos_5p, self.bam, contig_lengths)

            for read in reads:

                if (read.reference_start < pos_3p - self.insertion_spanning_length and
                    read.reference_end > pos_5p + self.insertion_spanning_length + 1 and
                    not self.contains_large_insertion(read, pos_3p, pos_5p)):

                    spanning_read_names.add(read.query_name)

            spanning_reads.append(len(spanning_read_names))

        assigned_pairs['spanning_count'] = spanning_reads
        assigned_pairs = assigned_pairs.loc[:, self.get_header_list()]
        return assigned_pairs

    def get_reads_at_site(self, contig, pos_3p, pos_5p, bam, contig_lengths):

        reads = []

        start = pos_3p - 1
        end = pos_5p + 2

        if start < 0:
            start = 0
        if end > contig_lengths[contig]:
            end = contig_lengths[contig]

        for read in bam.fetch(contig, start, end):
            if self.passes_read_filters(read):
                reads.append(read)

        return reads


    def passes_read_filters(self, read):
        if read.mapping_quality < self.min_alignment_quality:
            return False
        elif not sctools.read_meets_min_alignment_inner_length(read, self.min_alignment_inner_length):
            return False
        else:
            return True


    def filter_junction_spanning(self, pairs):

        pairs['total'] = (pairs['total_count_5p'] + pairs['total_count_3p'])/2
        pairs['spanning_prop'] = pairs['spanning_count'] / pairs['total']

        prop_cutoff = self.max_junction_spanning_prop
        pairs = pairs.query('spanning_prop < @prop_cutoff')

        pairs = pairs[self.get_header_list()]

        return pairs


    def contains_large_insertion(self, read, pos_3p, pos_5p):


        large_insert_3p = self.identify_large_insertion_at_site(pos_3p, read)
        large_insert_5p = self.identify_large_insertion_at_site(pos_5p, read)

        if large_insert_3p or large_insert_5p:
            return True
        else:
            return False


    def identify_large_insertion_at_site(self, position, read):

        blocks = read.get_blocks()

        if len(blocks) > 1:
            for i in range(len(blocks)-1):
                block1, block2 = blocks[i], blocks[i+1]
                if block1[1] == block2[0]:

                    if position == block1[1]-1:
                        insertion_length = pysamtools.get_insertion_length(position, read)
                        if insertion_length >= self.large_insertion_cutoff:
                            return True
                        else:
                            return False
                    elif position == block2[0]:
                        insertion_length = pysamtools.get_insertion_length(position, read, reverse=True)
                        if insertion_length >= self.large_insertion_cutoff:
                            return True
                        else:
                            return False
                else:
                    return False

        return False


    def block_overlaps_site(self, block, position):

        if block[0] <= position and block[1] > position:
            return True
        else:
            return False



import warnings
warnings.filterwarnings("ignore")
import sys
from glob import glob
from snakemake import shell
import click
from os.path import isfile
import pysam
from mgefinder import misc


def index_genome(genome_path, silence=True):
    if silence:
        shell('bwa index {genome_path} &> /dev/null;'.format(genome_path=genome_path))
    else:
        shell('bwa index {genome_path}'.format(genome_path=genome_path))

    return genome_is_indexed(genome_path)

def genome_is_indexed(genome_path):

    amb = genome_path + '.amb'
    ann = genome_path + '.ann'
    bwt = genome_path + '.bwt'
    pac = genome_path + '.pac'
    sa = genome_path + '.sa'

    index_files = [amb, ann, bwt, pac, sa]

    indexed = True
    files = glob(genome_path+'*')
    for f in index_files:
        if f not in files:
            indexed = False

    return indexed

def align_to_genome_pe(fastq1, fastq2, genome_path, out_sam, threads=1, verbose=False):
    if verbose:
        command = "bwa mem -t {threads} {genome_path} {fastq1} {fastq2} > {out_sam}; ".format(
            genome_path=genome_path, fastq1=fastq1, fastq2=fastq2, out_sam=out_sam, threads=threads)
    else:
        command = "bwa mem -t {threads} {genome_path} {fastq1} {fastq2} 2> /dev/null 1> {out_sam}; ".format(
            genome_path=genome_path, fastq1=fastq1, fastq2=fastq2, out_sam=out_sam, threads=threads)

    shell(command)

    if isfile(out_sam):
        return True
    else:
        return False

def align_to_genome_se(fastq1, genome_path, out_sam, threads=1, verbose=False):
    if verbose:
        command = "bwa mem -t {threads} {genome_path} {fastq1} > {out_sam}; ".format(
            genome_path=genome_path, fastq1=fastq1, out_sam=out_sam, threads=threads)
    else:
        command = "bwa mem -t {threads} {genome_path} {fastq1} 2> /dev/null 1> {out_sam}; ".format(
            genome_path=genome_path, fastq1=fastq1, out_sam=out_sam, threads=threads)

    shell(command)

    if isfile(out_sam):
        return True
    else:
        return False

def align_to_genome_fasta_pe(fasta1, fasta2, genome_path, out_sam, threads=1, verbose=False, additional_flags=''):
    if verbose:
        command = "bwa mem -t {threads} {genome_path} {fastq1} {fastq2} {additional_flags} > {out_sam}; ".format(
            genome_path=genome_path, fastq1=fasta1, fastq2=fasta2, out_sam=out_sam, threads=threads,
            additional_flags=additional_flags)
    else:
        command = "bwa mem -t {threads} {genome_path} {fastq1} {fastq2} {additional_flags} 2> /dev/null 1> {out_sam}; ".format(
            genome_path=genome_path, fastq1=fasta1, fastq2=fasta2, out_sam=out_sam, threads=threads,
            additional_flags=additional_flags)

    shell(command)

    if isfile(out_sam):
        return True
    else:
        return False


def align_fasta_to_genome(fasta, genome_path, out_bam, threads=1, silence=True, additional_flags=''):

    if silence:
        command1 = "rm -f {out_bam}.sam {out_bam}.sam.tmp; " \
                  "bwa mem {additional_flags} -t {threads} {genome_path} {fasta} 2> /dev/null 1> {out_bam}.sam;"
        command2 = "samtools view -h -F 4 {out_bam}.sam.secseq 2> /dev/null 1> {out_bam}.sam.tmp; " \
                  "samtools sort {out_bam}.sam.tmp 2> /dev/null 1> {out_bam}; " \
                  "samtools index {out_bam} 2> /dev/null;" \
                  "rm -f {out_bam}.sam {out_bam}.sam.secseq {out_bam}.sam.tmp".format(
            genome_path=genome_path, fasta=fasta, out_bam=out_bam, threads=threads, additional_flags=additional_flags)

    else:
        command1 = "rm -f {out_bam}.sam {out_bam}.sam.tmp; " \
                  "bwa mem {additional_flags} -t {threads} {genome_path} {fasta} 1> {out_sam}.sam;"
        command2 = "samtools view -h -F 4 {out_bam}.sam.secseq 1> {out_bam}.sam.tmp; " \
                  "samtools sort {out_bam}.sam.tmp 1> {out_bam}; " \
                  "samtools index {out_bam};" \
                  "rm -f {out_bam}.sam {out_bam}.sam.secseq {out_bam}.sam.tmp".format(
            genome_path=genome_path, fasta=fasta, out_bam=out_bam, threads=threads, additional_flags=additional_flags)

    shell(command1)
    add_sequence_to_secondary_alignment('{out_bam}.sam'.format(out_bam=out_bam),
                                        '{out_bam}.sam.secseq'.format(out_bam=out_bam))
    shell(command2)

    if isfile(out_bam):
        return True
    else:
        return False

def add_sequence_to_secondary_alignment(sam_file_in, sam_file_out):
    outfile = open(sam_file_out, 'w')
    infile = pysam.AlignmentFile(sam_file_in, 'r')
    outfile = pysam.AlignmentFile(sam_file_out, "w", template=infile)

    current_seq = None
    current_seq_reverse = None
    for read in infile:
        if read.query_sequence is not None:
            current_seq = read.query_sequence
            current_seq_reverse = read.is_reverse
        else:
            if current_seq_reverse == read.is_reverse:
                read.query_sequence = current_seq
            else:
                read.query_sequence = misc.revcomp(current_seq)

        outfile.write(read)
    outfile.close()



# Adapted from code by Shubhadeep Roychowdhury available at https://towardsdatascience.com/implementing-a-trie-data-structure-in-python-in-less-than-100-lines-of-code-a877ea23c1a1

import warnings
warnings.filterwarnings("ignore")
from typing import Tuple
import sys
from collections import defaultdict

class TrieNode(object):
    """
    Our trie node implementation. Very basic. but does the job
    """

    def __init__(self, char: str, parent=None):
        self.char = char
        self.children = {}
        self.parent = parent
        self.word_count = 0
        self.counter = 1
        self.total_lifetime_children = 0
        self.qual = 0


class Trie:

    def __init__(self):
        self.root = TrieNode('')
        self.total_words = 0
        self.total_nodes = 0
        self.total_base_quality = 0

    def add(self, word: str, quals: list):
        """
        Adding a word in the trie structure
        """

        if len(word) != len(quals):
            print("Added word and character quality vectors ahve different lengths")
            sys.exit()

        node = self.root
        for i in range(len(word)):
            char = word[i]
            qual = quals[i]
            found_in_child = False
            # Search for the character in the children of the present `node`

            if char in node.children:
                child = node.children[char]
                child.counter += 1
                child.qual += qual
                node = child
            else:
                new_node = TrieNode(char, node)
                new_node.qual += qual
                node.children[char] = new_node
                node.total_lifetime_children += 1
                node = new_node

        node.word_count += 1

    def load(self, word, quals, counts):
        """
        Loading a word in the trie structure
        """

        if len(word) != len(quals) != len(counts):
            print("Loading values have different lengths")
            sys.exit()

        node = self.root
        for i in range(len(word)):
            char = word[i]
            qual = quals[i]
            count = counts[i]
            found_in_child = False
            # Search for the character in the children of the present `node`

            if char in node.children:
                child = node.children[char]
                node = child
            else:
                new_node = TrieNode(char, node)
                new_node.qual = qual
                new_node.counter = count
                node.children[char] = new_node
                node.total_lifetime_children += 1
                node = new_node

    def make_subtrie(self, words):
        subtrie = Trie()

        for word in words:
            trie_node = self.root
            subtrie_node = subtrie.root

            for char in word:

                trie_child = trie_node.children[char]

                if char in subtrie_node.children:
                    subtrie_child = subtrie_node.children[char]

                else:
                    subtrie_child = TrieNode(char, subtrie_node)

                    subtrie_child.qual = trie_child.qual
                    subtrie_child.word_count = trie_child.word_count
                    subtrie.total_words += trie_child.word_count

                    subtrie_node.children[char] = subtrie_child
                    subtrie_node.total_lifetime_children = len(subtrie_node.children)


                trie_node = trie_child
                subtrie_node = subtrie_child


        return subtrie


    def load_words(self, words, quals, counts):

        for i in range(len(words)):
            word, qual, count = words[i], quals[i], counts[i]
            self.load(word, qual, count)


    def find_prefix(self, prefix: str) -> Tuple[bool, int]:
        """
        Check and return
          1. If the prefix exists in any of the words we added so far
          2. If yes then how may words actually have the prefix
        """
        node = self.root
        # If the root node has no children, then return False.
        # Because it means we are trying to search in an empty trie
        if not self.root.children:
            return False, 0
        for char in prefix:
            char_not_found = True
            # Search through all the children of the present `node`
            for child in node.children:
                if child.char == char:
                    # We found the char existing in the child.
                    char_not_found = False
                    # Assign node as the child containing the char and break
                    node = child
                    break
            # Return False anyway when we did not find a char.
            if char_not_found:
                return False, 0
        # Well, we are here means we have found the prefix. Return true to indicate that
        # And also the counter of the last node. This indicates how many words have this
        # prefix
        return True, node.counter

    def traverse_seqs(self, prefix='', node=None):
        if node is None:
            node = self.root

        word = prefix + node.char

        if len(node.children) == 0:
            return [word]

        words = []
        for child in node.children.values():
            words = words + self.traverse_seqs(word, child)

        return words

    def traverse_quals(self, prefix=[], node=None):
        if node is None:
            node = self.root
            qual = []
        else:
            qual = prefix + [node.qual]

        if len(node.children) == 0:
            return [qual]

        quals = []
        for child in node.children.values():
            quals = quals + self.traverse_quals(qual, child)

        return quals

    def traverse_counts(self, prefix=[], node=None):
        if node is None:
            node = self.root
            count = []
        else:
            count = prefix + [node.counter]

        if len(node.children) == 0:
            return [count]

        counts = []
        for child in node.children.values():
            counts = counts + self.traverse_counts(count, child)

        return counts

    def traverse_all(self):

        seqs = self.traverse_seqs()
        quals = self.traverse_quals()
        counts = self.traverse_counts()

        all = list(zip(seqs, quals, counts))
        return(all)

    def calc_total_words(self, word):
        total_words = 0
        node = self.root

        for c in word:
            total_words += node.children[c].word_count
            node = node.children[c]

        return total_words

    def calc_total_words_before_lifetime_child(self, word):
        total_words = 0
        node = self.root

        for c in word:
            if node.total_lifetime_children > 1:
                break
            total_words += node.children[c].word_count
            node = node.children[c]

        return total_words

    def calc_total_shared_words(self, word1, word2):

        shared_words = 0
        node = self.root

        for i in range(min([len(word1), len(word2)])):
            c1, c2 = word1[i], word2[i]
            if c1 != c2:
                break
            else:
                shared_words += node.children[c1].word_count

            node = node.children[c1]

        return shared_words

    def calc_total_unique_shared_words(self, word1, word2):

        shared_words = 0
        node = self.root

        for i in range(min([len(word1), len(word2)])):
            c1, c2 = word1[i], word2[i]
            if c1 != c2:
                break
            elif len(node.children) > 1:
                shared_words = node.children[c1].word_count
            else:
                shared_words += node.children[c1].word_count

            node = node.children[c1]

        return shared_words

    def calc_total_unique_words(self, word1, word2):

        unique_words = 0
        node = self.root

        same_path = True
        for i in range(len(word1)):
            c1 = word1[i]
            if i >= len(word2) or c1 != word2[i]:
                same_path = False

            if not same_path:
                unique_words += node.children[c1].word_count

            node = node.children[c1]
        return unique_words

    def delete_word(self, word):
        node = self.root
        for char in word:
            node = node.children[char]

        while True:
            parent_node = node.parent

            if parent_node.char == '' or len(parent_node.children) > 1:
                del parent_node.children[char]
                break

            del parent_node.children[char]
            node = parent_node
            char = node.char

    def calc_word_count_diff(self, word1, word2):

        node = self.root

    def make_consensus_word(self, mincount):

        total_count = sys.maxsize
        current_position_nodes = set([self.root])

        consensus = ''
        while len(current_position_nodes) > 0:

            total_count = 0
            next_position_nodes = set()
            charquals = defaultdict(int)

            # Add up char qualities and counts
            for n in current_position_nodes:
                charquals[n.char] += n.qual
                total_count += n.counter
                if n == self.root:
                    total_count = sys.maxsize

            if total_count < mincount:
                break

            # Get the next consensus character
            best_chars = self.get_best_qual_chars(charquals)
            if len(best_chars) > 1:
                consensus += 'N'
            else:
                consensus += best_chars.pop()

            # Load up the next level of nodes
            for n in current_position_nodes:
                for char in n.children:
                    next_position_nodes.add(n.children[char])

            current_position_nodes = next_position_nodes

        return consensus

    def get_best_qual_chars(self, charquals):
        bestchar = set([''])
        bestqual = -sys.maxsize
        for c in charquals:
            if charquals[c] == bestqual:
                bestchar.add(c)
            elif charquals[c] > bestqual:
                bestchar = set([c])
                bestqual = charquals[c]
        return bestchar

if __name__ == "__main__":
    trie = Trie()
    trie.add('hellothere', [40]*len('hellothere'))
    trie.add('hellothen', [40]*len('hellothen'))
    trie.add('hellohowareyou', [40]*len('hellohowareyou'))
    trie.add('hello', [40]*len('hello'))

    print(trie.calc_total_shared_words('hellothere', 'hellothen'))
    print(trie.calc_total_unique_words('hellothere', 'hellothen'))
    print(trie.calc_total_unique_words('hellothen', 'hellothere'))
    print(trie.calc_total_words_before_lifetime_child('hellothen'))


import sys
import warnings
warnings.filterwarnings("ignore")
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")
import pandas as pd
from mgefinder import bowtie2tools
from mgefinder.inferseq import InferSequence
import click
from os.path import dirname


def _inferseq_reference(pairsfile, inferseq_reference, min_perc_identity, max_internal_softclip_prop,
                        max_inferseq_size, min_inferseq_size, keep_intermediate, output_file):

    pairs = pd.read_csv(pairsfile, sep='\t', keep_default_na=False, na_values=[
        '-1.#IND', '1.#QNAN', '1.#IND', '-1.#QNAN', '#N/A', 'N/A', '#NA', 'NULL', 'NaN', '-NaN', 'nan', '-nan'])


    handle_empty_pairsfile(pairs, output_file)

    index_genome(inferseq_reference)
    tmp_dir = dirname(output_file)

    inferer = InferSequence(
        pairs, inferseq_reference, min_perc_identity, max_internal_softclip_prop, max_inferseq_size,
        min_inferseq_size, keep_intermediate, 'inferred_reference', tmp_dir
    )

    inferred_sequences = inferer.infer_sequences()

    click.echo("Inferred sequences for %d pairs..." % len(set(list(inferred_sequences['pair_id']))))
    click.echo("Writing results to file %s..." % output_file)
    if pairs.shape[0] > 0:
        sample_id = list(pairs['sample'])[0]
        inferred_sequences.insert(0, 'sample', sample_id)
    else:
        inferred_sequences.insert(0, 'sample', None)

    inferred_sequences.to_csv(output_file, sep='\t', index=False)


def index_genome(inferseq_reference):
    if not bowtie2tools.genome_is_indexed(inferseq_reference):
        click.echo("Indexing inferseq reference genome...")
        bowtie2tools.index_genome(inferseq_reference)
    click.echo("Genome has been indexed...")


def handle_empty_pairsfile(pairs, output_file):
    if pairs.shape[0] == 0:
        outfile = pd.DataFrame(columns=['pair_id', 'method', 'loc', 'inferred_seq_length', 'inferred_seq'])

        if not output_file:
            output_file = 'mgefinder.inferseq_reference.tsv'

        outfile.to_csv(output_file, sep='\t', index=False)
        click.echo("Empty pairs file, exiting...")
        sys.exit()


if __name__ == '__main__':
    _inferseq_reference()


import pandas as pd
import click
from collections import defaultdict
from mgefinder import fastatools

verbose=True


def _makefasta(clusterseq, summarize_clusters, output_prefix):

    fasta_all_path = output_prefix + '.all_seqs.fna'
    fasta_repr_path = output_prefix + '.repr_seqs.fna'

    click.echo('Reading clusterseq...')
    clusterseq = pd.read_table(clusterseq)

    click.echo('Reading summarized clusters...')
    summarize_clusters = pd.read_table(summarize_clusters)

    click.echo('Creating fasta for all unique sequences...')
    make_all_unique_fasta(clusterseq, summarize_clusters, fasta_all_path)

    click.echo('Creating fasta for representative sequences...')
    make_repr_cluster_fasta(summarize_clusters, fasta_repr_path)


def make_repr_cluster_fasta(summarize_clusters, fasta_repr_path):

    sequences = []
    names = []
    for cluster, seqid, group, repr_seq in zip(
            summarize_clusters.cluster, summarize_clusters.repr_seqid, summarize_clusters.group,
            summarize_clusters.repr_seq
    ):
        name = '_'.join([cluster, group, seqid])
        names.append(name)
        sequences.append(repr_seq)

    fastatools.write_sequences_to_fasta(sequences, fasta_repr_path, names)


def make_all_unique_fasta(clusterseq, summarize_clusters, fasta_all_path):

    keep_clusters = set(list(summarize_clusters.cluster))
    unique_seqs = defaultdict(set)
    for cluster, group, seqid, seq in zip(
            clusterseq.cluster, clusterseq.group, clusterseq.seqid, clusterseq.inferred_seq):

        if cluster in keep_clusters:
            unique_seqs['_'.join([cluster, group, seqid])].add(seq)

    names = list(unique_seqs.keys())
    sequences = [list(unique_seqs[name])[0] for name in names]

    fastatools.write_sequences_to_fasta(sequences, fasta_all_path, names)



import warnings
warnings.filterwarnings("ignore")
from mgefinder.sctools import *

def get_left_softclipped_reads_at_site(bam_file, contig, left_site, get_quals=False, softclip_only=False):
    left_softclipped_reads = []
    left_softclipped_quals = []

    start = left_site + 1
    end = left_site + 2

    contig_len = contig_length(bam_file, contig)
    if end > contig_len:
        end = contig_len

    for pu in bam_file.pileup(contig, start, end, truncate=True):

        for pr in pu.pileups:
            read = pr.alignment
            if is_left_softclipped_lenient_at_site(read, contig, left_site):

                if softclip_only:
                    keep_seq = left_softclipped_sequence(read)
                else:
                    keep_seq = read.query_sequence
                    keep_quals = get_query_qualities_ascii(read, bam_file)

                left_softclipped_reads.append(keep_seq)
                left_softclipped_quals.append(keep_quals)

    if get_quals:
        return left_softclipped_reads, left_softclipped_quals
    else:
        return left_softclipped_reads



def get_right_softclipped_reads_at_site(bam_file, contig, right_site, get_quals=False, softclip_only=False):
    right_softclipped_reads = []
    right_softclipped_quals = []

    start = right_site - 1
    end = right_site

    if start < 0:
        start = 0

    for pu in bam_file.pileup(contig, start, end, truncate=True):

        for pr in pu.pileups:
            read = pr.alignment
            if is_right_softclipped_lenient_at_site(read, contig, right_site):
                if softclip_only:
                    keep_seq = right_softclipped_sequence(read)
                else:
                    keep_seq = read.query_sequence
                    keep_quals = get_query_qualities_ascii(read, bam_file)

                right_softclipped_reads.append(keep_seq)
                right_softclipped_quals.append(keep_quals)

    if get_quals:
        return right_softclipped_reads, right_softclipped_quals
    else:
        right_softclipped_reads


def get_right_unmapped_reads(bam_file, contig, right_site, get_quals=False, search_region_length=500):
    right_unmapped_reads = []
    right_unmapped_quals = []

    start = right_site - search_region_length
    end = right_site

    if start < 0:
        start = 0

    for read in bam_file.fetch(contig, start, end):

        if (not read.is_reverse) and read.mate_is_unmapped:
            right_unmapped_reads.append(read.get_tag('MT'))
            right_unmapped_quals.append(read.get_tag('MQ'))

    if get_quals:
        return right_unmapped_reads, right_unmapped_quals
    else:
        return right_unmapped_reads


def get_left_unmapped_reads(bam_file, contig, left_site, get_quals=False, search_region_length=500):
    left_unmapped_reads = []
    left_unmapped_quals = []

    start = left_site + 1
    end = left_site + 1 + search_region_length

    contig_len = contig_length(bam_file, contig)
    if end > contig_len:
        end = contig_len

    for read in bam_file.fetch(contig, start, end):

        if read.is_reverse and read.mate_is_unmapped:
            left_unmapped_reads.append(read.get_tag('MT'))
            left_unmapped_quals.append(read.get_tag('MQ'))

    if get_quals:
        return left_unmapped_reads, left_unmapped_quals
    else:
        return left_unmapped_reads


def contig_length(bam, contig):
    return dict(zip(bam.references, bam.lengths))[contig]

def count_runthrough_reads(bam, contig, site, min_qual=20, min_alignment_inner_length=21):

    if site  < 0 or site >= contig_length(bam, contig):
        return 0
    else:
        count = 0
        for read in bam.fetch(contig, site, site+1):

            if read.mapping_quality < min_qual:
                continue

            if not read_meets_min_alignment_inner_length(read, min_alignment_inner_length):
                continue

            count += 1

        return count

def count_softclipped_reads(bam, contig, site, min_qual=20, min_alignment_inner_length=21):

    if site  < 0 or site >= contig_length(bam, contig):
        return 0
    else:
        count = 0
        for read in bam.fetch(contig, site+1, site+2):

            if read.mapping_quality < min_qual:
                continue

            if not read_meets_min_alignment_inner_length(read, min_alignment_inner_length):
                continue

            if is_left_softclipped_lenient_at_site(read, contig, site):
                count += 1

        for read in bam.fetch(contig, site - 1, site):

            if read.mapping_quality < min_qual:
                continue

            if not read_meets_min_alignment_inner_length(read, min_alignment_inner_length):
                continue

            if is_right_softclipped_lenient_at_site(read, contig, site):
                count += 1


        return count


def get_query_qualities_ascii(read, bam):
    return read.tostring(bam).split('\t')[10]


def query_qualities_to_phred(quals):
    return [ord(q)-33 for q in quals]


def get_perc_identity(read):
    total = read.query_alignment_length
    identity = (total - read.get_tag('NM')) / total
    return identity


def get_bam_contig_dict(bam_file):
    contig_dict = {}
    for contig in bam_file.header['SQ']:
        contig_dict[contig['SN']] = contig['LN']
    return contig_dict

def get_insertion_length(position, read, reverse=False):
    refpos = read.get_reference_positions(full_length=True)

    insert_length = 0

    if not reverse:

        for pos in refpos[refpos.index(position)+1:]:

            if pos is not None:
                break
            else:
                insert_length += 1

    else:
        refpos = refpos[::-1]
        for pos in refpos[refpos.index(position)+1:]:

            if pos is not None:
                break
            else:
                insert_length += 1

    return insert_length



from snakemake import shell
from glob import glob
from os.path import isfile
import click

def index_genome(genome_path, silence=True):
    if silence:
        shell('bowtie2-build -o 0 -q {genome_path} {genome_path} 2> /dev/null;'.format(genome_path=genome_path))
    else:
        shell('bowtie2-build -o 0 {genome_path} {genome_path}'.format(genome_path=genome_path))

    return genome_is_indexed(genome_path)


def genome_is_indexed(genome_path):

    one = genome_path + '.1.bt2'
    two = genome_path + '.2.bt2'
    three = genome_path + '.3.bt2'
    four = genome_path + '.4.bt2'
    five = genome_path + '.rev.1.bt2'
    six = genome_path + '.rev.2.bt2'

    index_files = [one, two, three, four, five, six]

    indexed = True
    files = glob(genome_path+'*')
    for f in index_files:
        if f not in files:
            indexed = False

    return indexed


def align_fasta_to_genome(fasta, genome_path, out_bam, threads=1, silence=True, additional_flags=''):

    if silence:
        command = "rm -f {out_bam}.sam {out_bam}.sam.tmp; " \
                  "bowtie2 {additional_flags} --local -x {genome_path} -p {threads} -f -U {fasta} -S {out_bam}.sam 2> /dev/null; " \
                  "samtools view -h -q 1 {out_bam}.sam 2> /dev/null 1> {out_bam}.sam.tmp; " \
                  "samtools sort {out_bam}.sam.tmp 2> /dev/null 1> {out_bam}; " \
                  "samtools index {out_bam} 2> /dev/null;" \
                  "rm -f {out_bam}.sam {out_bam}.sam.tmp".format(
            genome_path=genome_path, fasta=fasta, out_bam=out_bam, threads=threads, additional_flags=additional_flags)

    else:
        command = "rm -f {out_bam}.sam {out_bam}.sam.tmp; " \
                  "bowtie2 {additional_flags} --local -x {genome_path} -p {threads} -f -U {fasta} -S {out_bam}.sam; " \
                  "samtools view -h -q 1 {out_bam}.sam > {out_bam}.sam.tmp; " \
                  "samtools sort {out_bam}.sam.tmp > {out_bam}; " \
                  "samtools index {out_bam};" \
                  "rm -f {out_bam}.sam {out_bam}.sam.tmp".format(
            genome_path=genome_path, fasta=fasta, out_bam=out_bam, threads=threads, additional_flags=additional_flags)



    if not silence:
        click.echo("Executing command: %s" % command)

    shell(command, read=silence)

    if isfile(out_bam):
        return True
    else:
        return False

def align_paired_fasta_to_genome(fasta1, fasta2, genome_path, out_bam, threads=1, silence=False, additional_flags=''):

    if silence:
        command = "rm -f {out_bam}.sam {out_bam}.sam.tmp; " \
                  "bowtie2 {additional_flags} --local -x {genome_path} -p {threads} -f -1 {fasta1} -2 {fasta2} -S {out_bam}.sam &> /dev/null; " \
                  "samtools view -h -q 1 {out_bam}.sam 2> /dev/null 1> {out_bam}.sam.tmp; " \
                  "samtools sort {out_bam}.sam.tmp 2> /dev/null 1> {out_bam}; " \
                  "samtools index {out_bam} 2> /dev/null;" \
                  "rm -f {out_bam}.sam {out_bam}.sam.tmp".format(
            genome_path=genome_path, fasta1=fasta1, fasta2=fasta2, out_bam=out_bam, threads=threads, additional_flags=additional_flags)

    else:
        command = "rm -f {out_bam}.sam {out_bam}.sam.tmp; " \
                  "bowtie2 {additional_flags} --local -x {genome_path} -p {threads} -f -1 {fasta1} -2 {fasta2} -S {out_bam}.sam; " \
                  "samtools view -h -q 1 {out_bam}.sam > {out_bam}.sam.tmp; " \
                  "samtools sort {out_bam}.sam.tmp > {out_bam}; " \
                  "samtools index {out_bam};" \
                  "rm -f {out_bam}.sam {out_bam}.sam.tmp".format(
            genome_path=genome_path, fasta1=fasta1, fasta2=fasta2, out_bam=out_bam, threads=threads, additional_flags=additional_flags)



    if not silence:
        click.echo("Executing command: %s" % command)

    shell(command, read=silence)

    if isfile(out_bam):
        return True
    else:
        return False


from setuptools import setup, find_packages

setup(
    name="mgefinder",
    version='1.0.3',
    description='A toolbox for identifying mobile genetic element (MGE) insertions from short-read sequencing data of bacterial isolates.',
    url='https://github.com/bhattlab/MGEfinder',
    author="Matt Durrant",
    author_email="mdurrant@stanford.edu",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'click',
        'pandas',
        'biopython',
        'pysam',
        'editdistance',
        'scipy',
        'networkx',
        'tqdm'
    ],
    zip_safe=False,
    entry_points = {
        'console_scripts': [
            'mgefinder = mgefinder.main:cli'
        ],
}
)



from Bio import SeqIO
from snakemake import shell
from mgefinder import misc

def cluster_100p(infile, outfile):

    seqdict1 = dict()
    for rec in SeqIO.parse(infile, 'fasta'):
        seq = str(rec.seq)

        if seq not in seqdict1:
            seqdict1[seq] = rec

    seqdict2 = dict()
    for seq in seqdict1:
        fwd = seq
        rev = misc.revcomp(seq)

        seq_key = tuple(sorted([fwd, rev]))

        if seq_key not in seqdict2:
            seqdict2[seq_key] = seqdict1[seq]

    with open(outfile, "w") as handle:
        SeqIO.write(seqdict2.values(), handle, "fasta")


def cluster_reciprocal_identity(infile, outfile, threads=1, memory=800, alignment_coverage=0.99, perc_identity=0.99):


    shell('cd-hit-est -g 1 -aL {cov} -aS {cov} -d 0 -c {perc_identity} -T {threads} '
          '-M {memory} -i {infile} -o {outfile}'.format(
        cov = alignment_coverage, perc_identity = perc_identity, threads=threads, memory=memory,
        infile = infile, outfile = outfile
    ))


def parse_cdhit_output_100p(unclustered_fasta, cdhit_output):


    final_cluster_mappings = dict()

    seqdict_100p = dict()
    for rec in SeqIO.parse(cdhit_output, 'fasta'):
        fwd = str(rec.seq)
        rev = misc.revcomp(fwd)

        seq_key = tuple(sorted([fwd, rev]))
        seqdict_100p[seq_key] = rec


    for rec in SeqIO.parse(unclustered_fasta, 'fasta'):
        fwd = str(rec.seq)
        rev = misc.revcomp(fwd)

        seq_key = tuple(sorted([fwd, rev]))

        cluster = seqdict_100p[seq_key].id

        final_cluster_mappings[rec.id] = cluster

    return final_cluster_mappings


def parse_cdhit_output_perc_identity(cdhit_output):


    final_cluster_mappings = dict()
    with open(cdhit_output+'.clstr') as infile:
        all_cluster_seqs = []
        repr_seq = None
        for l in infile:
            l = l.strip()
            if l.startswith('>'):
                for seq in all_cluster_seqs:
                    final_cluster_mappings[seq] = repr_seq
                all_cluster_seqs = []
                repr_seq = None
            else:
                seq = l.split()[2].strip('>').strip('.')
                all_cluster_seqs.append(seq)
                if l[-1] == '*':
                    repr_seq = seq

        for seq in all_cluster_seqs:
            final_cluster_mappings[seq] = repr_seq

    return final_cluster_mappings


import pandas as pd
from collections import defaultdict
import click
import numpy as np
from mgefinder import misc


def _summarize(clusterseq, genotypes, output_prefix):

    summary_clusters_path = output_prefix + '.clusters.tsv'
    summary_groups_path = output_prefix + '.groups.tsv'

    click.echo('Reading clusterseq...')
    clusterseq = pd.read_table(clusterseq)

    click.echo('Reading genotypes...')
    genotypes = pd.read_table(genotypes)


    # GET RID OF THIS WHEN YOU FIX THE CLUSTERING
    clusterseq = clusterseq[-clusterseq.cluster.str.startswith('UClust')]
    genotypes = genotypes[-genotypes.cluster.str.startswith('UClust')]


    click.echo("Summarizing clusters...")
    cluster_summary = summarize_clusters(clusterseq, genotypes)

    click.echo("Summarizing groups...")
    group_summary = summarize_groups(clusterseq, genotypes)

    click.echo("Writing results to files %s and %s" % (summary_clusters_path, summary_groups_path))
    cluster_summary.to_csv(summary_clusters_path, sep='\t', index=False)
    group_summary.to_csv(summary_groups_path, sep='\t', index=False)


def summarize_clusters(clusterseq, genotypes):

    clusters2groups = get_clusters2groups(clusterseq.cluster, clusterseq.group)
    num_unique_sites = get_num_unique_sites(genotypes.cluster,
                                            zip(genotypes.contig, genotypes.pos_5p, genotypes.pos_3p),
                                            genotypes.conf)
    conf_counts = get_conf_counts(genotypes.cluster, genotypes.conf)
    seq_lengths = get_seq_lengths(clusterseq.cluster, clusterseq.inferred_seq)
    repr_seqs = get_repr_seq(clusterseq.cluster, clusterseq.seqid, clusterseq.inferred_seq)
    sample_stats = get_sample_stats(genotypes.cluster, genotypes['sample'])

    df1 = pd.merge(clusters2groups, num_unique_sites, how='inner', on=['cluster']).drop_duplicates()
    df2 = pd.merge(df1, seq_lengths, how='inner', on=['cluster']).drop_duplicates()
    df3 = pd.merge(df2, sample_stats, how='inner', on=['cluster']).drop_duplicates()
    df4 = pd.merge(df3, conf_counts, how='inner', on=['cluster']).drop_duplicates()
    final = pd.merge(df4, repr_seqs, how='inner', on=['cluster']).drop_duplicates().sort_values(
        ['num_unique_sites_all'], ascending=False
    )

    return final


def summarize_groups(clusterseq, genotypes):

    num_unique_sites = get_num_unique_sites(genotypes.group, zip(genotypes.contig, genotypes.pos_5p, genotypes.pos_3p),
                                            genotypes.conf)
    conf_counts = get_conf_counts(genotypes.group, genotypes.conf)
    seq_lengths = get_seq_lengths(clusterseq.group, clusterseq.inferred_seq)
    repr_seqs = get_repr_seq(clusterseq.group, clusterseq.seqid, clusterseq.inferred_seq)
    sample_stats = get_sample_stats(genotypes.group, genotypes['sample'])
    repr_cluster2seq = get_repr_cluster2seq(clusterseq.cluster, clusterseq.inferred_seq)
    repr_seqs = (pd.merge(repr_cluster2seq, repr_seqs, how='inner', on='repr_seq')[[
        'cluster', 'repr_cluster', 'repr_seqid', 'repr_seq_length', 'repr_seq'
    ]])

    df1 = pd.merge(num_unique_sites, seq_lengths, how='inner', on=['cluster']).drop_duplicates()
    df2 = pd.merge(df1, sample_stats, how='inner', on=['cluster']).drop_duplicates()
    df3 = pd.merge(df2, conf_counts, how='inner', on=['cluster']).drop_duplicates()
    final = pd.merge(df3, repr_seqs, how='inner', on=['cluster']).drop_duplicates().sort_values(
        ['num_unique_sites_all'], ascending=False
    )

    final.columns = ['group'] + list(final.columns[1:])

    return final


def get_clusters2groups(clusters, groups, header=['cluster', 'group']):

    clusters2groups = list(zip(clusters, groups))
    clusters2groups = pd.DataFrame(clusters2groups, columns = header)

    return clusters2groups


def get_num_unique_sites(clusters, sites, conf, header=['cluster', 'num_unique_sites_all', 'num_unique_sites_unambig']):

    unique_sites_all = defaultdict(set)
    unique_sites_unambig = defaultdict(set)

    for cluster, site, conf in zip(clusters, sites, conf):

        ambig = False
        if conf == 'A':
            ambig = True

        for clust in cluster.split(';'):

            unique_sites_all[clust].add(site)

            if not ambig:
                unique_sites_unambig[clust].add(site)


    unique_sites_all = [(cluster, len(unique_sites_all[cluster])) for cluster in unique_sites_all]
    unique_sites_unambig = [(cluster, len(unique_sites_unambig[cluster])) for cluster in unique_sites_unambig]

    unique_sites_all = pd.DataFrame(unique_sites_all, columns=header[:-1])
    unique_sites_unambig = pd.DataFrame(unique_sites_unambig, columns=[header[0]] + [header[-1]])

    out = pd.merge(unique_sites_all, unique_sites_unambig, how='left', on='cluster').fillna(0)

    return out


def get_conf_counts(clusters, confs, header=['cluster', 'IAwFC', 'IAwHC', 'IO', 'IAwoC', 'IDB',
                                             'ArSC', 'ArMS', 'ArML', 'A']):

    conf_counts = defaultdict(lambda: {lab: 0 for lab in header[1:]})

    for cluster, conf in zip(clusters, confs):

        for clust in cluster.split(';'):
            conf_counts[clust][conf] += 1

    conf_counts = [tuple([cluster] + [conf_counts[cluster][lab] for lab in header[1:]]) for cluster in conf_counts]

    conf_counts = pd.DataFrame(conf_counts, columns=header)

    return conf_counts


def get_seq_lengths(clusters, seqs, header=['cluster', 'num_unique_seqs', 'mean_length', 'min_length', 'max_length']):

    seq_lengths1 = defaultdict(set)

    for cluster, seq in zip(clusters, seqs):
        seq_lengths1[cluster].add(seq)

    seq_lengths2 = defaultdict(lambda: {'unique_seqs': set(), 'num_unique_seqs': 0,
                                        'mean_length': 0, 'min_length': 0, 'max_length': 0})
    for cluster in seq_lengths1:
        for seq in seq_lengths1[cluster]:
            seq_lengths2[cluster]['unique_seqs'].add(tuple(sorted([seq, misc.revcomp(seq)])))

    for cluster in seq_lengths2:

        all_seq_lengths = list(map(lambda x: len(x[0]), seq_lengths2[cluster]['unique_seqs']))
        seq_lengths2[cluster]['num_unique_seqs'] = len(all_seq_lengths)
        seq_lengths2[cluster]['mean_length'] = np.mean(all_seq_lengths)
        seq_lengths2[cluster]['min_length'] = np.min(all_seq_lengths)
        seq_lengths2[cluster]['max_length'] = np.max(all_seq_lengths)

    seq_lengths2 = [tuple([cluster] + [seq_lengths2[cluster][lab] for lab in header[1:]]) for cluster in seq_lengths2]
    seq_lengths2 = pd.DataFrame(seq_lengths2, columns=header)

    return seq_lengths2


def get_repr_seq(clusters, seqids, seqs, header=['cluster', 'repr_seqid', 'repr_seq_length', 'repr_seq']):

    seq_counts = defaultdict(lambda: defaultdict(int))

    for cluster, seqid, seq in zip(clusters, seqids, seqs):
        seq_counts[cluster][(seqid, seq)] += 1


    repr_seq = dict()
    for clust in seq_counts:

        max_seq = max(seq_counts[clust].items(), key=lambda x: x[1])

        repr_seq[clust] = max_seq[0]

    repr_seq = [(clust, repr_seq[clust][0], len(repr_seq[clust][1]), repr_seq[clust][1]) for clust in repr_seq]

    repr_seq = pd.DataFrame(repr_seq, columns=header)

    return repr_seq


def get_sample_stats(clusters, samples, header=['cluster', 'num_samples', 'mean_copy', 'max_copy']):

    sample_counts = defaultdict(lambda: defaultdict(int))

    for cluster, samp in zip(clusters, samples):

        for clust in cluster.split(';'):
            sample_counts[clust][samp] += 1


    sample_stats = defaultdict(dict)
    for cluster in sample_counts:

        counts = [x[1] for x in sample_counts[cluster].items()]
        sample_stats[cluster]['num_samples'] = len(sample_counts[cluster])
        sample_stats[cluster]['mean_copy'] = np.mean(counts)
        sample_stats[cluster]['max_copy'] = np.max(counts)

    sample_stats = [(cluster, sample_stats[cluster]['num_samples'],
                     sample_stats[cluster]['mean_copy'], sample_stats[cluster]['max_copy'])
                    for cluster in sample_stats]

    sample_stats = pd.DataFrame(sample_stats, columns=header)

    return sample_stats

def get_repr_cluster2seq(clusters, seqs, header=['repr_cluster', 'repr_seq']):

    repr_cluster2seq = set()

    for cluster, seq in zip(clusters, seqs):

        repr_cluster2seq.add((cluster, seq))

    repr_cluster2seq = pd.DataFrame(list(repr_cluster2seq), columns=header)

    return repr_cluster2seq


import warnings
warnings.filterwarnings("ignore")
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Alphabet import IUPAC
from mgefinder.misc import revcomp


def write_sequences_to_fasta(sequences, fasta, names=None):
    outseqs = []
    for i in range(len(sequences)):
        if names:
            name = names[i]
        else:
            name = i
        record = SeqRecord(Seq(sequences[i], IUPAC.IUPACAmbiguousDNA),
                           id=str(name), description=str(name))
        outseqs.append(record)
    with open(fasta, 'w') as output_handle:
        SeqIO.write(outseqs, output_handle, 'fasta')

def read_fasta(fasta):
    for rec in SeqIO.parse(fasta, 'fasta'):
        yield rec

def write_termini_to_fasta(termini, fasta):

    outseqs = []
    for index, row in termini.iterrows():
        name5p = str(index) + '_5p'
        name3p = str(index) + '_3p'
        seq5p = row['seq_5p']
        seq3p = row['seq_3p']

        record5p = SeqRecord(Seq(seq5p, IUPAC.IUPACAmbiguousDNA),
                           id=str(name5p), description=str(name5p))
        record3p = SeqRecord(Seq(seq3p, IUPAC.IUPACAmbiguousDNA),
                             id=str(name3p), description=str(name3p))

        outseqs.append(record5p)
        outseqs.append(record3p)

    with open(fasta, 'w') as output_handle:
        SeqIO.write(outseqs, output_handle, 'fasta')

def write_termini_to_paired_fasta(pairs, fasta_prefix):

    fiveprime_outseqs = []
    threeprime_outseqs = []
    for p in pairs:
        name5p = p['pair_id']
        name3p = p['pair_id']
        seq5p = p['seq_5p']
        seq3p = p['seq_3p']

        record5p = SeqRecord(Seq(seq5p, IUPAC.IUPACAmbiguousDNA),
                           id=str(name5p), description=str(name5p)+'_5p')
        record3p = SeqRecord(Seq(revcomp(seq3p), IUPAC.IUPACAmbiguousDNA),
                             id=str(name3p), description=str(name3p)+'_3p')

        fiveprime_outseqs.append(record5p)
        threeprime_outseqs.append(record3p)

    with open(fasta_prefix+'.R1.fasta', 'w') as output_handle:
        SeqIO.write(fiveprime_outseqs, output_handle, 'fasta')

    with open(fasta_prefix + '.R2.fasta', 'w') as output_handle:
        SeqIO.write(threeprime_outseqs, output_handle, 'fasta')

def write_termini_to_unpaired_fasta(pairs, fasta_prefix):

    fiveprime_outseqs = []
    threeprime_outseqs = []
    for p in pairs:
        name5p = p['pair_id']+'_1'
        name3p = p['pair_id']+'_2'
        seq5p = p['seq_5p']
        seq3p = p['seq_3p']

        record5p = SeqRecord(Seq(seq5p, IUPAC.IUPACAmbiguousDNA),
                           id=str(name5p), description=str(name5p)+'_5p')
        record3p = SeqRecord(Seq(revcomp(seq3p), IUPAC.IUPACAmbiguousDNA),
                             id=str(name3p), description=str(name3p)+'_3p')

        fiveprime_outseqs.append(record5p)
        threeprime_outseqs.append(record3p)

    with open(fasta_prefix+'.fasta', 'w') as output_handle:
        SeqIO.write(fiveprime_outseqs, output_handle, 'fasta')

    with open(fasta_prefix + '.fasta', 'a') as output_handle:
        SeqIO.write(threeprime_outseqs, output_handle, 'fasta')

def write_panisa_termini_to_fasta(termini, fasta):

    outseqs = []
    for index, row in termini.iterrows():
        name5p = str(index) + '_5p'
        name3p = str(index) + '_3p'
        seq5p = row['Left sequence']
        seq3p = row['Right sequence']

        record5p = SeqRecord(Seq(seq5p, IUPAC.IUPACAmbiguousDNA),
                           id=str(name5p), description=str(name5p))
        record3p = SeqRecord(Seq(seq3p, IUPAC.IUPACAmbiguousDNA),
                             id=str(name3p), description=str(name3p))

        outseqs.append(record5p)
        outseqs.append(record3p)

    with open(fasta, 'w') as output_handle:
        SeqIO.write(outseqs, output_handle, 'fasta')


from mgefinder import bwatools
from mgefinder import bowtie2tools
from mgefinder import samtools
from mgefinder.find import _find
from mgefinder.pair import _pair
from mgefinder.inferseqassembly import _inferseq_assembly
from Bio import SeqIO
from snakemake import shell
import click


def _wholegenome(reference, query, read_length, read_depth, min_alignment_quality, max_direct_repeat_length,
                 large_insertion_cutoff, query_id, output_prefix):

    min_alignment_inner_length = max_direct_repeat_length + 1

    if not bwatools.genome_is_indexed(reference):
        click.echo("Indexing reference...")
        bwatools.index_genome(reference)
    else:
        click.echo("Reference already indexed...")

    if not bowtie2tools.genome_is_indexed(query):
        click.echo("Indexing query...")
        bowtie2tools.index_genome(query)
    else:
        click.echo("Query already indexed...")

    click.echo("Making query reads...")
    query_reads_path = output_prefix + '.query.tmp.fq'
    make_reads(query, query_reads_path, read_length, read_depth)

    click.echo("Aligning query reads to reference...")
    out_sam = output_prefix + '.query.reference.tmp.sam'
    bwatools.align_to_genome_se(query_reads_path, reference, out_sam, threads=1, verbose=False)

    click.echo("Sorting and indexing alignment file...")
    out_bam = output_prefix + '.query.reference.tmp.bam'
    samtools.sort_coordinate(out_sam, out_bam, delete_in_bam=True)
    samtools.index(out_bam)

    find_file = output_prefix + '.find.tsv'
    _find(out_bam, min_softclip_length=8, min_softclip_count=1, min_alignment_quality=min_alignment_quality,
          min_alignment_inner_length=min_alignment_inner_length, min_distance_to_mate=max_direct_repeat_length + 2,
          min_softclip_ratio=0.01, max_indel_ratio=0.0, large_insertion_cutoff=large_insertion_cutoff,
          min_count_consensus=1, sample_id=query_id, output_file=find_file)

    pair_file = output_prefix + '.pair.tsv'
    _pair(find_file, out_bam, reference, max_direct_repeat_length=max_direct_repeat_length,
          min_alignment_quality=min_alignment_quality, min_alignment_inner_length=min_alignment_inner_length,
          max_junction_spanning_prop=0.01, large_insertion_cutoff=large_insertion_cutoff, output_file=pair_file)

    inferseq_file = output_prefix + '.inferseq.tsv'
    _inferseq_assembly(pair_file, out_bam, query, reference, min_perc_identity=0.95,
                       max_internal_softclip_prop=0.01, max_inferseq_size=500000,
                       min_inferseq_size=30, keep_intermediate=False, output_file=inferseq_file)

    shell('rm %s %s %s' % (query_reads_path, out_bam, out_bam+'.bai'))


def make_reads(genome, out_fastq, readlength, depth):

    with open(out_fastq, 'w') as out:
        step = int(readlength / depth)
        recnum = 0
        for rec in SeqIO.parse(genome, 'fasta'):
            for i in range(0, len(rec.seq)-readlength, step):
                recnum += 1
                outread = rec.seq[i:i + readlength]
                out.write('@SEQ' + str(recnum) + '\n')
                out.write(str(outread) + '\n+\n')
                out.write('F'*len(outread) + '\n')
    return out_fastq







import warnings
warnings.filterwarnings("ignore")
import sys
from mgefinder.workflow import _workflow
from mgefinder.find import _find
from mgefinder.pair import _pair
from mgefinder.inferseqassembly import _inferseq_assembly
from mgefinder.inferseqoverlap import _inferseq_overlap
from mgefinder.inferseqreference import _inferseq_reference
from mgefinder.inferseqdatabase import _inferseq_database
from mgefinder.makedatabase import _makedatabase
from mgefinder.formatbam import _formatbam
from mgefinder.recall import _recall
from mgefinder.help import CustomHelp
from mgefinder.clusterseq import _clusterseq
from mgefinder.genotype import _genotype
from mgefinder.summarize import _summarize
from mgefinder.makefasta import _makefasta
from mgefinder.wholegenome import _wholegenome
from mgefinder.dependencies import check_dependencies
from mgefinder.misc import aligned_bwa, BWACheckError

import click
from os.path import join, dirname, abspath
from shutil import copyfile

WORKFLOW_SNAKEFILE = join(dirname(__file__), 'workflow/Snakefile')
WORKFLOW_CONFIG = join(dirname(__file__), 'workflow/config.yml')

check_dependencies()

@click.group(cls=CustomHelp)
def cli():
    """Command-line tools to identify mobile genetic element insertions from short-read sequencing data."""
    pass

@cli.command(short_help='Run mgefinder on a working directory using a snakemake workflow.', help_priority=1)
@click.argument('workdir', type=click.Path(exists=True))
@click.option('--snakefile', '-s', default=WORKFLOW_SNAKEFILE, help="The Snakefile file to use when running the snakemake workflow. Default is provided with mgefinder installation.")
@click.option('--configfile', '-c', default=WORKFLOW_CONFIG, help="The config file to use when running the snakemake workflow. Default is provided with mgefinder installation.")
@click.option('--cores', '-t', default=1, help="The number of processors to run while finding flank extensions. default=1")
@click.option('--memory', '-m', default=16000, help="Memory limit in megabytes. default=16000; 0 for unlimited")
@click.option('--unlock/--no-unlock',  default=False, help="Unlock working directory if necessary.")
@click.option('--rerun-incomplete/--no-rerun-incomplete',  default=False, help="Rerun incomplete files in the workflow.")
@click.option('--keep-going/--no-keep-going',  default=False, help="Keep going with independent jobs if one fails.")
def workflow(workdir, snakefile, configfile, cores, memory, unlock, rerun_incomplete, keep_going):
    """A click access point for the workflow module. This is used for creating the command line interface."""

    log_params(command='workflow', workdir=workdir, snakefile=snakefile, configfile=configfile, cores=cores, memory=memory,
                 unlock=unlock, rerun_incomplete=rerun_incomplete, keep_going=keep_going)
    _workflow(workdir, snakefile, configfile, cores, memory, unlock, rerun_incomplete, keep_going)


@cli.command(short_help='Get copies of the default Snakefile and config files.', help_priority=2)
def getworkflow():
    """A click access point for the getworkflow module. This is used for creating the command line interface."""

    click.echo("Copying Snakefile and config file to current working directory...")
    copyfile(WORKFLOW_SNAKEFILE, "Snakefile")
    copyfile(WORKFLOW_CONFIG, "config.yml")
    click.echo("Done.")



@cli.command(short_help='Find insertion sites and reconstruct termini of inserted sequence', help_priority=3)
@click.argument('bamfile', type=click.Path(exists=True))
@click.option('--min_softclip_length', '-minlen', default=8, help="For a softclipped site to be considered, there must be at least one softclipped read of this length. default=8")
@click.option('--min_softclip_count', '-mincount', default=2, help="For a softclipped site to be considered, there must be at least this many softclipped reads at the site. default=2")
@click.option('--min_alignment_quality', '-minq', default=20, help="For a read to be considered, it must meet this alignment quality cutoff. default=20")
@click.option('--min_alignment_inner_length', '-minial', default=21, help="If a read is softclipped on both ends, the aligned portion must be at least this long. Ideally, set this equal to 1 + max_direct_repeat_length. default=21")
@click.option('--min_distance_to_mate', '-mindist', default=22, help="A minimum distance to a potential nearby mate, filters out sites that have no pairs. default=22")
@click.option('--min_softclip_ratio', '-minratio', default=0.10, help="For a softclipped site to be considered, the proportion of softclipped sites + large insertions must not fall below this value. default=0.10")
@click.option('--max_indel_ratio', '-maxir', default=0.03, help="For a softclipped site to be considered, the proportion of small insertions/deletions at this site must not be above this value. default=0.03")
@click.option('--large_insertion_cutoff', '-lins', default=30, help="Keep large insertions if they meet this length. default=30")
@click.option('--min_count_consensus', '-mcc', default=2, help="When building the consensus sequence, stop building consensus if read count drops below this cutoff. default=2")
@click.option('--sample_id', '-id', default=None, help="Specify an ID to use for this sample, default is the absolute path to the bam file")
@click.option('--check-bwa/--no-check-bwa', default=True, help="Check to ensure that BWA MEM was used to generate BAM file. default=True")
@click.option('--output_file', '-o', default='mgefinder.find.tsv', help="The output file to save the results. default=mgefinder.find.tsv")
def find(bamfile, min_softclip_length, min_softclip_count, min_alignment_quality, min_alignment_inner_length,
               min_distance_to_mate, min_softclip_ratio, max_indel_ratio, large_insertion_cutoff,
               min_count_consensus, sample_id, check_bwa, output_file):
    """A click access point for the find module. This is used for creating the command line interface."""

    if sample_id is None:
        sample_id = abspath(bamfile)

    if check_bwa and not aligned_bwa(bamfile):
        raise BWACheckError("ERROR: Please use BWA MEM to align your reads to the reference genome. Other aligners may not be compatible. To disable this check, use the --no-check-bwa flag.")
        sys.exit()
    

    log_params(command='find',
                 bamfile=bamfile, min_softclip_length=min_softclip_length, min_softclip_count=min_softclip_count,
                 min_alignment_quality=min_alignment_quality, min_alignment_inner_length=min_alignment_inner_length,
                 min_distance_to_mate=min_distance_to_mate, min_softclip_ratio=min_softclip_ratio,
                 max_indel_ratio=max_indel_ratio, large_insertion_cutoff=large_insertion_cutoff,
                 min_count_consensus=min_count_consensus, sample_id=sample_id, check_bwa=check_bwa, 
                 output_file=output_file)
    _find(bamfile, min_softclip_length, min_softclip_count, min_alignment_quality, min_alignment_inner_length,
                min_distance_to_mate, min_softclip_ratio, max_indel_ratio, large_insertion_cutoff,
                min_count_consensus, sample_id, output_file)


@cli.command(short_help="Pair identified termini with each other to represent 5' and 3' ends of inserted sequence.", help_priority=4)
@click.argument('findfile', type=click.Path(exists=True))
@click.argument('bamfile', type=click.Path(exists=True))
@click.argument('genome', type=click.Path(exists=True))
@click.option('--max_direct_repeat_length', '-maxdr', default=20, help="The maximum length of a direct repeat to consider a pair. default=20")
@click.option('--min_alignment_quality', '-minq', default=20, help="For a read to be considered, it must meet this alignment quality cutoff. default=20")
@click.option('--min_alignment_inner_length', '-minial', default=21, help="If a read is softclipped on both ends, the aligned portion must be at least this long. Ideally, set this equal to 1 + maximum direct repeat length. default=21")
@click.option('--max_junction_spanning_prop', '-maxjsp', default=0.15, help="Removes pairs where this proportion of readsextend across both insertion junctions without softclipping, an indication that the site is a duplicated region. default=0.15")
@click.option('--large_insertion_cutoff', '-lins', default=30, help="Keep large insertions if they meet this length. default=30")
@click.option('--output_file', '-o', default='mgefinder.pairs.tsv', help="The output file to save the results. default=mgefinder.pairs.tsv")
def pair(findfile, bamfile, genome, max_direct_repeat_length, min_alignment_quality,
               min_alignment_inner_length, max_junction_spanning_prop, large_insertion_cutoff, output_file=None):

    log_params(command='pair', findfile=findfile, bamfile=bamfile, genome=genome,
                 max_direct_repeat_length=max_direct_repeat_length, min_alignment_quality=min_alignment_quality,
                 min_alignment_inner_length=min_alignment_inner_length,
                 max_junction_spanning_prop=max_junction_spanning_prop, large_insertion_cutoff=large_insertion_cutoff,
                 output_file=output_file)
    _pair(findfile, bamfile, genome, max_direct_repeat_length, min_alignment_quality,
          min_alignment_inner_length, max_junction_spanning_prop, large_insertion_cutoff, output_file)


@cli.command(short_help='Infers the identity of an inserted sequence by aligning terminus pairs to an assembled genome.', help_priority=5)
@click.argument('pairsfile', type=click.Path(exists=True))
@click.argument('bamfile', type=click.Path(exists=True))
@click.argument('inferseq_assembly', type=click.Path(exists=True))
@click.argument('inferseq_reference', type=click.Path(exists=True))
@click.option('--min_perc_identity', '-minident', default=0.95, help="Only consider matches with a percentage identity above this threshold. default=0.95")
@click.option('--max_internal_softclip_prop', '-maxclip', default=0.05, help="Do not consider matches with internal softclipped ends exceeding this proportion of the total read. default=0.05")
@click.option('--max_inferseq_size', '-maxsize', default=200000, help="Do not consider inferred sequences over this size. default=200000")
@click.option('--min_inferseq_size', '-minsize', default=30, help="Do not consider inferred sequences below this size. default=30")
@click.option('--keep-intermediate/--no-keep-intermediate', default=False, help="Keep intermediate files. default=False")
@click.option('--output_file', '-o', default='mgefinder.inferseq_assembly.tsv', help="The output file to save the results. default=mgefinder.inferseq_assembly.tsv")
def inferseq_assembly(pairsfile, bamfile, inferseq_assembly, inferseq_reference, min_perc_identity,
                      max_internal_softclip_prop, max_inferseq_size, min_inferseq_size, keep_intermediate, output_file=None):
    """Infers the identity of an inserted sequence by aligning terminus pairs to an assembled genome."""

    log_params(command='inferseq_assembly', pairsfile=pairsfile, bamfile=bamfile, inferseq_assembly=inferseq_assembly,
                 inferseq_reference=inferseq_reference, min_perc_identity=min_perc_identity,
                      max_internal_softclip_prop=max_internal_softclip_prop, max_inferseq_size=max_inferseq_size,
                 min_inferseq_size=min_inferseq_size, keep_intermediate=keep_intermediate, output_file=output_file)
    _inferseq_assembly(pairsfile, bamfile, inferseq_assembly, inferseq_reference, min_perc_identity,
                       max_internal_softclip_prop, max_inferseq_size, min_inferseq_size, keep_intermediate, output_file)


@cli.command(short_help='Infers the identity of an inserted sequence by aligning terminus pairs to a reference genome. Ideal for re-sequencing experiments where evolved strains are closely related to the reference genome used.',
             help_priority =6)
@click.argument('pairsfile', type=click.Path(exists=True))
@click.argument('inferseq_reference', type=click.Path(exists=True))
@click.option('--min_perc_identity', '-minident', default=0.95, help="Only consider matches with a percentage identity above this threshold. default=0.95")
@click.option('--max_internal_softclip_prop', '-maxclip', default=0.05, help="Do not consider matches with internal softclipped ends exceeding this proportion of the total read. default=0.05")
@click.option('--max_inferseq_size', '-maxsize', default=200000, help="Do not consider inferred sequences over this size. default=200000")
@click.option('--min_inferseq_size', '-minsize', default=30, help="Do not consider inferred sequences below this size. default=30")
@click.option('--keep-intermediate/--no-keep-intermediate', default=False, help="Keep intermediate files. default=False")
@click.option('--output_file', '-o', default='mgefinder.inferseq_reference.tsv', help="The output file to save the results. default=mgefinder.inferseq_reference.tsv")
def inferseq_reference(pairsfile, inferseq_reference, min_perc_identity, max_internal_softclip_prop,
                       max_inferseq_size, min_inferseq_size, keep_intermediate, output_file=None):
    """
    Infers the identity of an inserted sequence by aligning terminus pairs to a reference genome.
    Ideal for re-sequencing experiments where evolved strains are closely related to the reference genome used.
    """
    log_params(command='inferseq_reference', pairsfile=pairsfile, inferseq_reference=inferseq_reference,
                 min_perc_identity=min_perc_identity, max_internal_softclip_prop=max_internal_softclip_prop,
                 max_inferseq_size=max_inferseq_size, min_inferseq_size=min_inferseq_size,
                 keep_intermediate=keep_intermediate, output_file=output_file)
    _inferseq_reference(pairsfile, inferseq_reference, min_perc_identity, max_internal_softclip_prop,
                        max_inferseq_size, min_inferseq_size, keep_intermediate, output_file)


@cli.command(short_help='Infers the identity of an inserted sequence by checking if they overlap with one another. Only identifies an inserted sequence if the consensus termini are long enough to span the entire insertion.',
             help_priority=7)
@click.argument('pairsfile', type=click.Path(exists=True))
@click.option('--min_overlap_score', '-minscore', default=10, help="The minimum overlap score to keep inferred sequence. default=10")
@click.option('--min_overlap_perc_identity', '-minopi', default=0.9, help="The minimum overlap percent identity to keep inferred sequence. default=0.9")
@click.option('--min_inferseq_size', '-minsize', default=30, help="Do not consider inferred sequences below this size. default=30")
@click.option('--output_file', '-o', default='mgefinder.inferseq_overlap.tsv', help="The output file to save the results. default=mgefinder.inferseq_overlap.tsv")
def inferseq_overlap(pairsfile, min_overlap_score, min_overlap_perc_identity, min_inferseq_size, output_file=None):
    """
    Infers the identity of an inserted sequence by checking if they overlap with one another.
    Only identifies an inserted sequence if the consensus termini are long enough to span the entire insertion.
    """
    log_params(command='inferseq_overlap', pairsfile=pairsfile, min_overlap_score=min_overlap_score,
                 min_overlap_perc_identity=min_overlap_perc_identity, min_inferseq_size=min_inferseq_size,
                 output_file=output_file)
    _inferseq_overlap(pairsfile, min_overlap_score, min_overlap_perc_identity, min_inferseq_size, output_file)


@cli.command(short_help="Make database from inferred sequences.", help_priority=8)
@click.argument('inferseqfiles', type=click.Path(exists=True), nargs=-1)
@click.option('--minimum_size', '-minsize', default=30, help="The minimum size of inferred sequence to use in the database. default=30")
@click.option('--maximum_size', '-maxsize', default=200000, help="The maximum size of inferred sequence to use in the database. default=200000")
@click.option('--threads', '-t', default=1, help="The number of processors to use clustering sequences. default=1")
@click.option('--memory', '-m', default=16000, help="Memory limit in megabytes. default=16000; 0 for unlimited")
@click.option('--force/--no-force', default=False, help="Force overwriting of output directory.")
@click.option('--output_dir', '-o', default='mgefinder.database', help="The output directory to save the results. default=mgefinder.database")
@click.option('--prefix', '-p', default='mgefinder.database', help="The prefix used to name the database files. default=mgefinder.database")
def makedatabase(inferseqfiles, minimum_size, maximum_size, threads, memory, force, output_dir=None, prefix=None):
    """
    Makes a FASTA database of all elements identified in provided inferseq files.
    """
    log_params(command='makedatabase', inferseqfiles=inferseqfiles, minimum_size=minimum_size,
                 maximum_size=maximum_size, threads=threads, memory=memory, force=force, output_dir=output_dir,
                 prefix=prefix)
    _makedatabase(inferseqfiles, minimum_size, maximum_size, threads, memory, force, output_dir, prefix)


@cli.command(short_help='Infers the identity of an inserted sequence by aligning flank pairs to an database of known inserted elements.', help_priority=9)
@click.argument('pairsfile', type=click.Path(exists=True))
@click.argument('inferseq_database', type=click.Path(exists=True))
@click.option('--min_perc_identity', '-minident', default=0.90, help="Only consider matches with a percentage identity above this threshold. default=0.90")
@click.option('--max_internal_softclip_prop', '-maxclip', default=0.05, help="Do not consider matches with internal softclipped ends exceeding this proportion of the total read. default=0.05")
@click.option('--max_edge_distance', '-maxedgedist', default=10, help="Reads must align within this number of bases from the edge of an element to be considered. default=10")
@click.option('--output_file', '-o', default='mgefinder.inferseq_database.tsv', help="The output file to save the results. default=mgefinder.inferseq_database.tsv")
@click.option('--keep-intermediate/--no-keep-intermediate', default=False, help="Keep intermediate files. default=False")
def inferseq_database(pairsfile, inferseq_database, min_perc_identity,  max_internal_softclip_prop, max_edge_distance, output_file=None, keep_intermediate=False):
    """Infers the identity of an inserted sequence by aligning flank pairs to a database of known inserted elements."""
    log_params(command='inferseq_database', pairsfile=pairsfile, inferseq_database=inferseq_database, 
                 min_perc_identity=min_perc_identity, max_internal_softclip_prop=max_internal_softclip_prop, 
                 max_edge_distance=max_edge_distance, output_file=output_file, keep_intermediate=keep_intermediate)
    _inferseq_database(pairsfile, inferseq_database, min_perc_identity, max_internal_softclip_prop, max_edge_distance, output_file, keep_intermediate)


@cli.command(short_help="Cluster inferred sequences.", help_priority=10)
@click.argument('inferseqfiles', type=click.Path(exists=True), nargs=-1)
@click.option('--minimum_size', '-minsize', default=70, help="The minimum size of inferred sequence to use in the database. default=70")
@click.option('--maximum_size', '-maxsize', default=200000, help="The maximum size of inferred sequence to use in the database. default=200000")
@click.option('--threads', '-t', default=1, help="The number of processors to use clustering sequences. default=1")
@click.option('--memory', '-m', default=16000, help="Memory limit in megabytes. default=16000; 0 for unlimited")
@click.option('--output_file', '-o', default='metamgefinder.clusterseq.tsv', help="The output file to save the results. default=metamgefinder.clusterseq.tsv")
def clusterseq(inferseqfiles, minimum_size, maximum_size, threads, memory, output_file=None):
    """Clusters elements identified in inferseq files using CD-HIT-EST"""
    log_params(command='clusterseq', inferseqfiles=inferseqfiles, minimum_size=minimum_size, maximum_size=maximum_size,
               threads=threads, memory=memory, output_file=output_file)
    _clusterseq(inferseqfiles, minimum_size, maximum_size, threads, memory, output_file)



@cli.command(short_help="Assign final genotypes to samples.", help_priority=11)
@click.argument('clusterseq', type=click.Path(exists=True), nargs=1)
@click.argument('pairfiles', type=click.Path(exists=True), nargs=-1)
@click.option('--filter-clusters-inferred-assembly/--no-filter-clusters-inferred-assembly', default=True, help="Removes sequence clusters that were never inferred from an assembly or by overlap if True. default=True")
@click.option('--output_file', '-o', default='mgefinder.genotype.tsv', help="The output file to save the results. default=mgefinder.genotype.tsv")
def genotype(clusterseq, pairfiles, filter_clusters_inferred_assembly, output_file):
    """Genotypes MGE insertions of input samples with respect to the reference genome."""
    log_params(command='genotype', clusterseq=clusterseq, pairfiles=pairfiles,
               filter_clusters_inferred_assembly=filter_clusters_inferred_assembly,
               output_file=output_file)
    _genotype(clusterseq, pairfiles, filter_clusters_inferred_assembly, output_file)


@cli.command(short_help="Summarize the clusters identified by mgefinder.", help_priority=12)
@click.argument('clusterseq', type=click.Path(exists=True), nargs=1)
@click.argument('genotypes', type=click.Path(exists=True), nargs=1)
@click.option('--output_prefix', '-o', default='mgefinder.summarize', help="The output file to save the results. default=mgefinder.summarize")
def summarize(clusterseq, genotypes, output_prefix):
    """Summarizes the findings of MGEfinder workflow"""
    log_params(command='summarize', clusterseq=clusterseq, genotypes=genotypes, output_prefix=output_prefix)
    _summarize(clusterseq, genotypes,  output_prefix)


@cli.command(short_help="Make FASTA files of the identified sequence clusters.", help_priority=13)
@click.argument('clusterseq', type=click.Path(exists=True), nargs=1)
@click.argument('summarize_clusters', type=click.Path(exists=True), nargs=1)
@click.option('--output_prefix', '-o', default='mgefinder', help="The output file to save the results. default=mgefinder")
def makefasta(clusterseq, summarize_clusters, output_prefix):
    """Makes FASTA files of the putative integrative MGEs identified in the pipeline"""
    log_params(command='makefasta', clusterseq=clusterseq, summarize_clusters=summarize_clusters,
               output_prefix=output_prefix)
    _makefasta(clusterseq, summarize_clusters,  output_prefix)


@cli.command(short_help="Formats a BAM file for use with mgefinder. Usually not necessary, unless using the experiment extendpairs command.", help_priority=14)
@click.argument('in_sam', type=click.Path(exists=True))
@click.argument('out_bam')
@click.option('--single-end', is_flag=True, default=False, help="Add this flag for single-end files. default=False")
@click.option('--keep-tmp-files', is_flag=True, default=False, help="Add this flag if you want to keep intermediate temporary files. default=False")
def formatbam(in_sam, out_bam, single_end, keep_tmp_files):
    """Formats a BAM file for use with mgefinder. Usually not necessary, unless using the experiment extendpairs command."""
    log_params(command='formatbam', in_sam=in_sam, out_bam=out_bam, single_end=single_end, keep_tmp_files=keep_tmp_files)
    _formatbam(in_sam, out_bam, single_end, keep_tmp_files)



@cli.command(short_help='Recall softclip counts and runthrough counts from BAM file at specified insertion sites', help_priority=15)
@click.argument('pairsfile', type=click.Path(exists=True))
@click.argument('bamfile', type=click.Path(exists=True))
@click.option('--min_alignment_quality', '-minq', default=20, help="For a read to be considered, it must meet this alignment quality cutoff. default=20")
@click.option('--min_alignment_inner_length', '-minial', default=21, help="If a read is softclipped on both ends, the aligned portion must be at least this long. Ideally, set this equal to 1 + maximum direct repeat length. default=21")
@click.option('--large_insertion_cutoff', '-lins', default=30, help="Same parameter as that used for find command. default=30")
@click.option('--output_file', '-o', default='mgefinder.recall.tsv', help="The output file to save results to. default=mgefinder.recall.tsv")
def recall(pairsfile, bamfile, min_alignment_quality, min_alignment_inner_length, large_insertion_cutoff, output_file):
    """Recall softclip counts and runthrough counts from BAM file at specified insertions sites."""
    log_params(command='recall', pairsfile=pairsfile, bamfile=bamfile, min_alignment_quality=min_alignment_quality,
               min_alignment_inner_length=min_alignment_inner_length, large_insertion_cutoff=large_insertion_cutoff,
               output_file=output_file)
    _recall(pairsfile, bamfile, min_alignment_quality, min_alignment_inner_length, large_insertion_cutoff, output_file)

@cli.command(short_help='Identify insertions using whole-genome alignments of complete genomes.', help_priority=16)
@click.argument('reference', type=click.Path(exists=True))
@click.argument('query', type=click.Path(exists=True))
@click.option('--read_length', '-rl', default=500, help="The length of the reads to simulate. Sets an effective minimum size for identified insertions. default=100")
@click.option('--read_depth', '-rd', default=5, help="The depth of coverage produced to compare genomes. default=5")
@click.option('--min_alignment_quality', '-minq', default=20, help="For a read to be considered, it must meet this alignment quality cutoff. default=20")
@click.option('--max_direct_repeat_length', '-maxdr', default=20, help="Only call insertions where the predicted direct repeat length is at most {max_direct_repeat_length}. default=20.")
@click.option('--large_insertion_cutoff', '-lins', default=30, help="Same parameter as that used for find command. default=30") 
@click.option('--query_id', '-id', default=None, help="Specify an ID to use for this query, default is the absolute path to the query file.")
@click.option('--output_prefix', '-o', default='mgefinder.wholegenome', help="The prefix of saved output files. default=mgefinder.wholegenomev")
def wholegenome(reference, query, read_length, read_depth, min_alignment_quality, max_direct_repeat_length, large_insertion_cutoff, query_id, output_prefix):
    """Identify insertions using whole-genome alignments of complete genomes. Inputs can be draft assemblies or completed genomes"""
    if query_id is None:
        query_id = abspath(query)

    log_params(command='wholegenome', reference=reference, query=query,
               read_length=read_length, read_depth=read_depth,
               min_alignment_quality=min_alignment_quality,
               max_direct_repeat_length=max_direct_repeat_length,
               large_insertion_cutoff=large_insertion_cutoff,
               query_id=query_id, output_prefix=output_prefix)
    _wholegenome(reference, query, read_length, read_depth, min_alignment_quality,
                 max_direct_repeat_length, large_insertion_cutoff, query_id, output_prefix)


def log_params(**kwargs):
    click.echo("#### PARAMETERS ####")
    click.echo('\n'.join(list(map(lambda x: ': '.join(list(map(str, x))), kwargs.items()))))
    click.echo("####################")

if __name__ == '__main__':

    cli()



import warnings
warnings.filterwarnings("ignore")
import pandas as pd
from collections import defaultdict
import click
from os.path import isfile
import sys

def _genotype(clusterseq, pairfiles, filter_clusters_inferred_assembly, output_file):


    click.echo("Loading clusterseq file...")
    clusterseq = pd.read_table(clusterseq)[['sample', 'pair_id', 'method', 'seqid', 'cluster', 'group']]

    click.echo("Parsing pair files")
    if len(pairfiles) == 1 and is_path_list(pairfiles[0]):
        pairfiles = [l.strip() for l in open(pairfiles[0], 'r')]

    click.echo("Loading pair files...")
    pairs = combine_pair_files(pairfiles)


    genotyper = Genotyper(clusterseq, pairs, filter_clusters_inferred_assembly)


    if pairs.shape[0] == 0:
        click.echo("No termini found in the input file...")

        genotypes = genotyper.get_header_dataframe()

        if output_file:
            genotypes.to_csv(output_file, sep='\t', index=False)


    else:

        genotypes = genotyper.genotype()

        if output_file:
            click.echo("Saving results to file %s" % output_file)
            genotypes.to_csv(output_file, sep='\t', index=False)

        return genotypes


def combine_pair_files(pairfiles):

    keep_header = ['sample', 'pair_id', 'contig', 'pos_5p', 'pos_3p']

    click.echo('Loading file {num1}/{num2}: {f}'.format(num1=1, num2=len(pairfiles), f=pairfiles[0]))

    f0 = pairfiles[0]
    pairs = pd.read_table(f0)[keep_header]


    for i, f in enumerate(pairfiles[1:]):
        click.echo('Loading file {num1}/{num2}: {f}'.format(num1=i + 2, num2=len(pairfiles), f=f))

        df = pd.read_table(f)[keep_header]
        pairs = pairs.append(df)

    return pairs


def is_path_list(f):
    with open(f) as infile:
        for l in infile:

            if len(l.strip().split()) > 1:
                return False

            if not isfile(l.strip()):
                return False

    return True

class Genotyper:
    clusterseq = None
    pairs = None
    filter_clusters_inferred_assembly = None

    def __init__(self, clusterseq, pairs, filter_clusters_inferred_assembly=True):

        self.clusterseq = clusterseq
        self.pairs = pairs
        self.filter_clusters_inferred_assembly = filter_clusters_inferred_assembly

        self.pairs['pair_id'] = list(map(int, list(self.pairs['pair_id'])))
        self.pairs['pos_5p'] = list(map(int, list(self.pairs['pos_5p'])))
        self.pairs['pos_3p'] = list(map(int, list(self.pairs['pos_3p'])))
        self.clusterseq['pair_id'] = list(map(int, list(self.clusterseq['pair_id'])))

    def genotype(self):

        if self.filter_clusters_inferred_assembly:
            click.echo("Filtering out clusters that are never inferred from an assembly...")
            self.clusterseq = self.apply_filter_clusters_inferred_assembly(self.clusterseq)

        combined = pd.merge(self.pairs, self.clusterseq, how='left', on=['sample', 'pair_id'])

        no_inference = combined[combined.method.isnull()]
        has_inference = combined[combined.method.notnull()]

        self.report_null_genotypes(combined, no_inference, has_inference)
        genotypes = self.assign_genotypes_heuristic(has_inference)
        genotypes = self.resolve_ambiguous_genotypes(genotypes)

        return genotypes

    def apply_filter_clusters_inferred_assembly(self, clusterseq):

        method_counts = defaultdict(lambda: defaultdict(int))
        for cluster, method in zip(clusterseq.cluster, clusterseq.method):
            method_counts[cluster][method] += 1

        keep_clusters = set([clust for clust in method_counts
                             if method_counts[clust]['inferred_assembly_with_full_context'] > 0 or
                             method_counts[clust]['inferred_assembly_with_half_context'] > 0 or
                             method_counts[clust]['inferred_assembly_without_context'] > 0])

        exclude_clusters = set([clust for clust in set(clusterseq.cluster) if clust not in keep_clusters])

        click.echo(
            "Excluding %d clusters that were only inferred from the reference genome..." % len(exclude_clusters))

        clusterseq = clusterseq[[clust in keep_clusters for clust in clusterseq.cluster]]

        return clusterseq

    def report_null_genotypes(self, combined, no_inference, has_inference):

        total_count = combined[['sample', 'pair_id']].drop_duplicates().shape[0]
        total_no_inference = no_inference[['sample', 'pair_id']].drop_duplicates().shape[0]
        total_has_inference = has_inference[['sample', 'pair_id']].drop_duplicates().shape[0]

        click.echo(
            "Out of %d candidate insertions, %d had some inferred identity, while %d had no inferred identity." %
            (total_count, total_has_inference, total_no_inference)
        )

    def assign_genotypes_heuristic(self, data):

        click.echo("Assigning initial genotypes...")

        # IAwFC
        IAwFC = data[data.method == 'inferred_assembly_with_full_context']
        IAwFC['conf'] = 'IAwFC'
        IAwFC = self.combine_unresolved_seqs(IAwFC.drop(columns='method'))

        genotyped = IAwFC[['sample', 'pair_id']].drop_duplicates()
        genotyped['pair_id'] = pd.to_numeric(genotyped['pair_id'])

        # IAwHC
        IAwHC = pd.merge(data, genotyped, on=['sample', 'pair_id'], how='outer', indicator=True)
        IAwHC = IAwHC[IAwHC._merge == 'left_only']
        IAwHC = IAwHC.drop(columns='_merge')
        IAwHC = IAwHC[IAwHC.method == 'inferred_assembly_with_half_context']
        IAwHC['conf'] = 'IAwHC'
        IAwHC = self.combine_unresolved_seqs(IAwHC.drop(columns='method'))

        genotyped = pd.concat([genotyped, IAwHC[['sample', 'pair_id']]]).drop_duplicates()
        genotyped['pair_id'] = pd.to_numeric(genotyped['pair_id'])

        # IO
        IO = pd.merge(data, genotyped, on=['sample', 'pair_id'], how='outer', indicator=True)
        IO = IO[IO._merge == 'left_only']
        IO = IO.drop(columns='_merge')
        IO = IO[IO.method == 'inferred_overlap']
        IO['conf'] = 'IO'
        IO = self.combine_unresolved_seqs(IO.drop(columns='method'))

        genotyped = pd.concat([genotyped, IO[['sample', 'pair_id']]]).drop_duplicates()
        genotyped['pair_id'] = pd.to_numeric(genotyped['pair_id'])

        # IAwoC
        IAwoC = pd.merge(data, genotyped, on=['sample', 'pair_id'], how='outer', indicator=True)
        IAwoC = IAwoC[IAwoC._merge == 'left_only']
        IAwoC = IAwoC.drop(columns='_merge')
        IAwoC = IAwoC[IAwoC.method == 'inferred_assembly_without_context']
        IAwoC['conf'] = 'IAwoC'
        IAwoC = self.combine_unresolved_seqs(IAwoC.drop(columns='method'))

        genotyped = pd.concat([genotyped, IAwoC[['sample', 'pair_id']]]).drop_duplicates()
        genotyped['pair_id'] = pd.to_numeric(genotyped['pair_id'])

        # IDB
        IDB = pd.merge(data, genotyped, on=['sample', 'pair_id'], how='outer', indicator=True)
        IDB = IDB[IDB._merge == 'left_only']
        IDB = IDB.drop(columns='_merge')
        IDB = IDB[(IDB.method == 'inferred_database') | (IDB.method == 'inferred_reference')]
        IDB['conf'] = 'IDB'
        IDB = self.combine_unresolved_seqs(IDB.drop(columns='method'))

        out = IAwFC.append(IAwHC).append(IO).append(IAwoC).append(IDB).drop_duplicates()
        out['pair_id'] = pd.to_numeric(out['pair_id'])

        return out

    def resolve_ambiguous_genotypes(self, data):

        click.echo("Identifying ambiguous genotypes...")

        counts = data.groupby(['sample', 'pair_id']).size().rename('n').reset_index()

        data = pd.merge(data, counts, on=['sample', 'pair_id'], how='inner')
        data['ambiguous'] = data.n > 1
        data = data.drop(columns='n')
        

        ambiguous = data[data.ambiguous == True]
        not_ambiguous = data[data.ambiguous == False]

        if ambiguous.shape[0] == 0:
            return not_ambiguous

        click.echo("Resolving ambiguous genotypes where possible...")

        mobile_clusters = self.calculate_mobile_clusters(data)
        cluster_counts_per_site = self.count_clusters_per_site(data)

        unresolved = ambiguous.copy()

        resolved_all_sample_comparison, unresolved = self.resolve_all_sample_comparison(
            unresolved, cluster_counts_per_site
        )


        resolved_mobile, unresolved = self.resolve_mobile(unresolved, mobile_clusters)

        combined_unresolved = self.combine_unresolved_clusters(unresolved)

        out = pd.concat([not_ambiguous, resolved_all_sample_comparison,
                         resolved_mobile, combined_unresolved], sort=False)[self.get_header_list()]

        return out

    def combine_unresolved_seqs(self, unresolved):

        combined = defaultdict(set)

        for sample, pair_id, contig, pos_5p, pos_3p, group, conf, cluster, seqid in zip(
                unresolved['sample'], unresolved.pair_id, unresolved.contig, unresolved.pos_5p, unresolved.pos_3p,
                unresolved.group, unresolved.conf, unresolved.cluster, unresolved.seqid):

            combined[(sample, pair_id, contig, pos_5p, pos_3p, cluster, group, conf)].add(seqid)

        combined = [list(key) + [';'.join(sorted(list(combined[key])))] for key in combined.keys()]

        combined = pd.DataFrame(
            combined, columns=['sample', 'pair_id', 'contig', 'pos_5p', 'pos_3p', 'cluster', 'group', 'conf', 'seqid']
        )

        combined = combined[self.get_header_list()]

        return combined

    def combine_unresolved_clusters(self, unresolved):

        combined = defaultdict(lambda: [set(), set()])

        for sample, pair_id, contig, pos_5p, pos_3p, group, conf, cluster, seqid in zip(
                unresolved['sample'], unresolved.pair_id, unresolved.contig, unresolved.pos_5p, unresolved.pos_3p,
                unresolved.group, unresolved.conf, unresolved.cluster, unresolved.seqid):

            combined[(sample, pair_id, contig, pos_5p, pos_3p, group, conf)][0].add(seqid)
            combined[(sample, pair_id, contig, pos_5p, pos_3p, group, conf)][1].add(cluster)

        combined = [list(key) + [';'.join(sorted(list(combined[key][0])))] + [';'.join(sorted(list(combined[key][1])))]
                    for key in combined.keys()]

        combined = pd.DataFrame(
            combined, columns=['sample', 'pair_id', 'contig', 'pos_5p', 'pos_3p', 'group', 'conf', 'seqid', 'cluster']
        )

        combined = combined[self.get_header_list()]
        combined['conf'] = 'A'

        return combined


    def calculate_mobile_clusters(self, data):

        mobile_lenient = defaultdict(set)
        for contig, pos_3p, pos_5p, cluster in zip(data.contig, data.pos_3p, data.pos_5p, data.cluster):
            mobile_lenient[cluster].add((contig, pos_3p, pos_5p))

        high_conf = set(['IAwFC', 'IAwHC', 'IO', 'IAwoC'])
        mobile_strict = defaultdict(set)
        for contig, pos_3p, pos_5p, cluster, conf, in zip(data.contig, data.pos_3p, data.pos_5p, data.cluster, data.conf):
            if conf in high_conf:
                mobile_strict[cluster].add((contig, pos_3p, pos_5p))

        mobile_lenient = [(cluster, len(mobile_lenient[cluster])) for cluster in mobile_lenient]
        mobile_lenient = pd.DataFrame(mobile_lenient, columns=['cluster', 'mobile_lenient'])
        mobile_lenient['mobile_lenient'] = [True if i > 2 else False for i in list(mobile_lenient['mobile_lenient'])]

        mobile_strict = [(cluster, len(mobile_strict[cluster])) for cluster in mobile_strict]
        mobile_strict = pd.DataFrame(mobile_strict, columns=['cluster', 'mobile_strict'])
        mobile_strict['mobile_strict'] = [True if i > 2 else False for i in list(mobile_strict['mobile_strict'])]

        mobile = pd.merge(mobile_lenient, mobile_strict, how='left', on='cluster').fillna(0)

        return mobile

    def count_clusters_per_site(self, data):

        cluster_counts = defaultdict(int)
        for contig, pos_5p, pos_3p, cluster in zip(data.contig, data.pos_5p, data.pos_3p, data.cluster):
            cluster_counts[(contig, pos_5p, pos_3p, cluster)] += 1

        cluster_counts = [list(group) + [cluster_counts[group]] for group in cluster_counts]

        cluster_counts = pd.DataFrame(cluster_counts, columns=['contig', 'pos_5p', 'pos_3p', 'cluster', 'n'])

        return cluster_counts

    def resolve_all_sample_comparison(self, unresolved, cluster_counts):

        if unresolved.shape[0] == 0:
            return pd.DataFrame(columns=list(unresolved.columns)), unresolved

        resolved = (pd.merge(unresolved, cluster_counts, how='inner', on=['contig', 'pos_5p', 'pos_3p', 'cluster']).
                    groupby(['sample', 'pair_id']).
                    apply(lambda g: g[g['n'] == g['n'].max()]).
                    reset_index(drop=True).
                    groupby(['sample', 'pair_id']).
                    filter(lambda g: len(g) == 1).
                    assign(ambiguous=lambda x: False).
                    assign(conf=lambda x: 'ArSC').
                    drop(columns='n'))

        unresolved = (pd.merge(unresolved, cluster_counts, how='inner', on=['contig', 'pos_5p', 'pos_3p', 'cluster']).
                      groupby(['sample', 'pair_id']).
                      apply(lambda g: g[g['n'] == g['n'].max()]).
                      reset_index(drop=True).
                      groupby(['sample', 'pair_id']).
                      filter(lambda g: len(g) > 1).
                      drop(columns='n').reset_index(drop=True))

        return resolved, unresolved


    def resolve_mobile(self, unresolved, mobile_clusters):

        if unresolved.shape[0] == 0:
            return pd.DataFrame(columns=list(unresolved.columns)), unresolved

        # Resolve using mobile strict definition
        filtered = pd.merge(unresolved, mobile_clusters, how='inner', on='cluster')
        filtered = filtered[filtered.mobile_strict == True]

        if filtered.shape[0] == 0:
            return pd.DataFrame(columns=list(unresolved.columns)), unresolved

        resolved_mobile_strict = (filtered.
                    groupby(['sample', 'pair_id']).
                    filter(lambda g: len(g) == 1).
                    assign(ambiguous=lambda x: False).
                    assign(conf=lambda x: 'ArMS').
                    drop(columns=['mobile_strict', 'mobile_lenient']))


        # Resolve using mobile lenient definition
        filtered = pd.merge(unresolved, mobile_clusters, how='inner', on='cluster')
        filtered = filtered[filtered.mobile_lenient == False]

        filtered = pd.merge(filtered, resolved_mobile_strict[['sample', 'pair_id']], how='outer',
                            on=['sample', 'pair_id'], indicator=True)
        filtered = filtered[filtered._merge == 'left_only'].drop(columns='_merge')


        if filtered.shape[0] == 0:
            return pd.DataFrame(columns=list(unresolved.columns)), unresolved

        resolved_mobile_lenient = (filtered.
                                   groupby(['sample', 'pair_id']).
                                   filter(lambda g: len(g) == 1).
                                   assign(ambiguous=lambda x: False).
                                   assign(conf=lambda x: 'ArML').
                                   drop(columns=['mobile_strict', 'mobile_lenient']))

        # Now get the unresolved
        all_resolved = pd.concat([resolved_mobile_strict, resolved_mobile_lenient]).drop_duplicates()
        merged = pd.merge(unresolved, all_resolved[['sample', 'pair_id']], on=['sample', 'pair_id'], how='outer', indicator=True)
        merged = merged[merged._merge == 'left_only']
        unresolved = merged.drop(columns='_merge').assign(conf=lambda x: 'A')

        
        return all_resolved, unresolved

    def get_header_list(self):
        header = ['sample', 'pair_id', 'contig', 'pos_3p', 'pos_5p', 'seqid', 'cluster', 'group', 'conf']

        return header

    def get_header_dataframe(self):
        terminus_pairs = pd.DataFrame(columns=self.get_header_list())

        return terminus_pairs


import warnings
import pysam
warnings.filterwarnings("ignore")

from scipy.stats import poisson
from bisect import bisect_left, bisect_right


def revcomp(read):
    reversed_seq = ''
    for l in reversed(read.upper()):
        if l == 'A':
            reversed_seq += 'T'
        elif l == 'T':
            reversed_seq += 'A'
        elif l == 'C':
            reversed_seq += 'G'
        elif l == 'G':
            reversed_seq += 'C'
        else:
            reversed_seq += l
    return reversed_seq


def poisson_test_greater(x, mu):

    return 1 - poisson.cdf(x-1, mu)


def takeClosestSmaller(sorted_list, number):

    if len(sorted_list) == 0:
        return None

    pos = bisect_left(sorted_list, number)

    if pos == 0:
        if sorted_list[0] >= number:
            return None
        else:
            return sorted_list[0]

    if pos == len(sorted_list):
        if sorted_list[-1] < number:
            return sorted_list[-1]
        else:
            return sorted_list[-2]

    if sorted_list[pos] < number:
        return sorted_list[pos]
    else:
        return sorted_list[pos-1]


def takeClosestLarger(sorted_list, number):

    if len(sorted_list) == 0:
        return None

    pos = bisect_right(sorted_list, number)

    if pos == len(sorted_list):
        if sorted_list[-1] <= number:
            return None
        else:
            return sorted_list[-1]

    if pos == 0:
        if sorted_list[0] > number:
            return sorted_list[0]
        else:
            return sorted_list[1]

    if sorted_list[pos] > number:
        return sorted_list[pos]
    else:
        return sorted_list[pos+1]



if __name__ == "__main__":
    print(takeClosestSmaller([], 100), 100)
    print()
    print(takeClosestSmaller([1], 0), 0)
    print(takeClosestSmaller([1], 1), 1)
    print(takeClosestSmaller([1], 2), 2)
    print()
    print(takeClosestSmaller([1, 2], 0), 0)
    print(takeClosestSmaller([1, 2], 1), 1)
    print(takeClosestSmaller([1, 2], 2), 2)
    print(takeClosestSmaller([1, 2], 3), 3)
    print()
    print(takeClosestSmaller([1, 2, 3], 0), 0)
    print(takeClosestSmaller([1, 2, 3], 1), 1)
    print(takeClosestSmaller([1, 2, 3], 2), 2)
    print(takeClosestSmaller([1, 2, 3], 3), 3)
    print(takeClosestSmaller([1, 2, 3], 4), 4)
    print(takeClosestSmaller([1, 2, 3], 5), 5)
    print()
    print()
    print(takeClosestLarger([], 100), 100)
    print()
    print(takeClosestLarger([1], 0), 0)
    print(takeClosestLarger([1], 1), 1)
    print(takeClosestLarger([1], 2), 2)
    print()
    print(takeClosestLarger([1, 2], 0), 0)
    print(takeClosestLarger([1, 2], 1), 1)
    print(takeClosestLarger([1, 2], 2), 2)
    print(takeClosestLarger([1, 2], 3), 3)
    print()
    print(takeClosestLarger([1, 2, 3], 0), 0)
    print(takeClosestLarger([1, 2, 3], 1), 1)
    print(takeClosestLarger([1, 2, 3], 2), 2)
    print(takeClosestLarger([1, 2, 3], 3), 3)
    print(takeClosestLarger([1, 2, 3], 4), 4)
    print(takeClosestLarger([1, 2, 3], 5), 5)

def aligned_bwa(bampath):
    bam = pysam.AlignmentFile(bampath)
    header = str(bam.header)
    try:
        header.index('ID:bwa')
        return True
    except ValueError:
        return False

def BWACheckError(Exception):
    pass

